import time
from threading import Timer
from typing import Self

from blueapi.client import BlueapiClient
from blueapi.client.event_bus import AnyEvent
from blueapi.service.model import TaskRequest
from blueapi.worker import ProgressEvent
from dodal.utils import get_beamline_name
from pydantic import BaseModel

from mx_bluesky.common.external_interaction.alerting import (
    Metadata,
    get_alerting_service,
)
from mx_bluesky.common.utils.log import LOGGER


class TaskMonitor:
    """
    Implements a context manager that on entry sets a timer for the BlueAPI Task to be executed within the with-block.
    The body should register on_blueapi_event as an event handler in the call to blueapi_client.run_task().

    If the timer expires before task completion, we either:
       * raise an error alert and cancel the task
       * raise an alert and reset the timer
    depending on whether received events indicate the task is waiting on a long-running event e.g. waiting for beam,
    or whether it is unexpectedly stuck.
    """

    DEFAULT_TIMEOUT_S = 600
    FEEDBACK_STATUS_NAME = "xbpm_feedback-pos_stable"

    def __init__(self, blueapi_client: BlueapiClient, task_request: TaskRequest):
        self._task_request = task_request
        self._alerting_service = get_alerting_service()
        self._blueapi_client = blueapi_client
        self._is_waiting_for_beam: bool = False
        self._timer = None

    def __enter__(self) -> Self:
        self._is_waiting_for_beam = False
        self._start_time = time.monotonic()
        self._reset_timer()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._timer:
            self._timer.cancel()

    def on_blueapi_event(self, event: AnyEvent):
        match event:
            case ProgressEvent(statuses=statuses):
                feedback_statuses = [
                    s
                    for s in statuses.values()
                    if s.display_name == self.FEEDBACK_STATUS_NAME
                ]
                if feedback_statuses:
                    feedback_status = feedback_statuses[0]
                    waiting_for_beam = feedback_status.current != feedback_status.target
                    if self._is_waiting_for_beam != waiting_for_beam:
                        self._reset_timer()
                        self._is_waiting_for_beam = waiting_for_beam
                    LOGGER.info(
                        f"Hyperion blueapi reports feedback status = {self._is_waiting_for_beam}"
                    )

    def on_timeout_expiry(self):
        if not self._is_waiting_for_beam:
            self._cancel_request()
            self._raise_alert_collection_is_stuck()
        else:
            self._raise_alert_collection_is_waiting_for_beam()
            self._reset_timer()

    def _reset_timer(self):
        if self._timer:
            self._timer.cancel()
        self._timer = Timer(self.DEFAULT_TIMEOUT_S, self.on_timeout_expiry)
        self._timer.start()

    def _raise_alert_collection_is_stuck(self):
        beamline = get_beamline_name("")
        self._alerting_service.raise_alert(
            f"UDC encountered an error on {beamline}",
            f"Hyperion Supervisor detected that BlueAPI was stuck for {self.DEFAULT_TIMEOUT_S} seconds.",
            self._extract_metadata(),
        )

    def _raise_alert_collection_is_waiting_for_beam(self):
        now = time.monotonic()
        minutes = int((now - self._start_time) // 60)
        beamline = get_beamline_name("")
        self._alerting_service.raise_alert(
            f"Hyperion is paused waiting for beam on {beamline}.",
            f"Hyperion has been paused waiting for beam for {minutes} minutes.",
            self._extract_metadata(),
        )

    def _extract_metadata(self):
        match self._task_request.params:
            case {"parameters": parameters}:
                if isinstance(parameters, BaseModel):
                    match parameters.model_dump():
                        case {
                            "sample_id": sample_id,
                            "visit": visit,
                            "sample_puck": container,
                        }:
                            return {
                                Metadata.SAMPLE_ID: sample_id,
                                Metadata.VISIT: visit,
                                Metadata.CONTAINER: container,
                            }
        return {}

    def _cancel_request(self):
        self._blueapi_client.abort(
            f"Hyperion Supervisor detected that BlueAPI was stuck for {self.DEFAULT_TIMEOUT_S} seconds."
        )



import bluesky.preprocessors as bpp
from dodal.common import inject

from mx_bluesky.beamlines.i24.jungfrau_commissioning.composites import (
    RotationScanComposite,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan import (
    ExternalRotationScanParams,
    rotation_scan_plan,
)


# Can remove this plan after https://github.com/DiamondLightSource/mx-daq-ui/issues/101
@bpp.run_decorator()
def gui_run_jf_rotation_scan(
    filename: str,
    exposure_time_s: float,
    omega_start_deg: float,
    omega_increment_deg: float,
    det_distance_mm: float,
    sample_id: int,
    transmissions: list[float],
    composite: RotationScanComposite = inject(),
):
    params = ExternalRotationScanParams(
        transmission_fractions=transmissions,
        exposure_time_s=exposure_time_s,
        omega_start_deg=omega_start_deg,
        rotation_increment_per_image_deg=omega_increment_deg,
        filename=filename,
        detector_distance_mm=det_distance_mm,
        sample_id=sample_id,
    )

    yield from rotation_scan_plan(composite, params)



import bluesky.plan_stubs as bps
import pydantic
from bluesky.utils import MsgGenerator
from dodal.beamlines.i03 import BL
from dodal.common.beamlines.beamline_parameters import (
    get_beamline_parameters,
)
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.collimation_table import CollimationTable
from dodal.devices.cryostream import (
    CryoStreamGantry,
    CryoStreamSelection,
    OxfordCryoJet,
    OxfordCryoStream,
)
from dodal.devices.cryostream import InOut as CryoInOut
from dodal.devices.fluorescence_detector_motion import FluorescenceDetector
from dodal.devices.fluorescence_detector_motion import InOut as FlouInOut
from dodal.devices.hutch_shutter import InterlockedHutchShutter, ShutterDemand
from dodal.devices.motors import XYZStage
from dodal.devices.mx_phase1.beamstop import BeamstopPositions
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.robot import BartRobot, PinMounted
from dodal.devices.scintillator import InOut as ScinInOut
from dodal.devices.scintillator import Scintillator
from dodal.devices.smargon import Smargon
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutterState

from mx_bluesky.common.device_setup_plans.robot_load_unload import robot_unload
from mx_bluesky.common.experiment_plans.beamstop_check import (
    BeamstopCheckDevices,
    move_beamstop_in_and_verify_using_diode,
)
from mx_bluesky.common.utils.exceptions import BeamlineCheckFailureError
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.external_interaction.config_server import (
    get_hyperion_feature_settings,
)
from mx_bluesky.hyperion.parameters.constants import CONST

_GROUP_PRE_BEAMSTOP_CHECK = "pre_beamstop_check"
_GROUP_POST_BEAMSTOP_CHECK = "post_beamstop_check"


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class UDCDefaultDevices(BeamstopCheckDevices):
    collimation_table: CollimationTable
    cryojet: OxfordCryoJet
    cryostream: OxfordCryoStream
    cryostream_gantry: CryoStreamGantry
    fluorescence_det_motion: FluorescenceDetector
    hutch_shutter: InterlockedHutchShutter
    lower_gonio: XYZStage
    robot: BartRobot
    scintillator: Scintillator
    gonio: Smargon
    oav: OAV


class UnexpectedSampleError(BeamlineCheckFailureError): ...


class CryoStreamError(BeamlineCheckFailureError): ...


def move_to_udc_default_state(devices: UDCDefaultDevices):
    """Moves beamline to known positions prior to UDC start"""
    yield from _verify_correct_cryostream_selected(devices.cryostream_gantry)

    yield from _check_cryostream(devices)

    # Close fast shutter before opening hutch shutter
    yield from bps.abs_set(devices.sample_shutter, ZebraShutterState.CLOSE, wait=True)

    commissioning_mode_enabled = yield from bps.rd(devices.baton.commissioning)

    if commissioning_mode_enabled:
        LOGGER.warning("Not opening hutch shutter - commissioning mode is enabled.")
    else:
        yield from bps.abs_set(
            devices.hutch_shutter, ShutterDemand.OPEN, group=_GROUP_PRE_BEAMSTOP_CHECK
        )

    yield from bps.abs_set(devices.scintillator.selected_pos, ScinInOut.OUT, wait=True)

    yield from bps.abs_set(
        devices.fluorescence_det_motion.pos,
        FlouInOut.OUT,
        group=_GROUP_PRE_BEAMSTOP_CHECK,
    )

    yield from bps.abs_set(
        devices.collimation_table.inboard_y,
        0,
        group=_GROUP_PRE_BEAMSTOP_CHECK,
    )
    yield from bps.abs_set(
        devices.collimation_table.outboard_y, 0, group=_GROUP_PRE_BEAMSTOP_CHECK
    )
    yield from bps.abs_set(
        devices.collimation_table.upstream_y, 0, group=_GROUP_PRE_BEAMSTOP_CHECK
    )
    yield from bps.abs_set(
        devices.collimation_table.upstream_x, 0, group=_GROUP_PRE_BEAMSTOP_CHECK
    )
    yield from bps.abs_set(
        devices.collimation_table.downstream_x, 0, group=_GROUP_PRE_BEAMSTOP_CHECK
    )

    # Wait for all of the above to complete
    yield from bps.wait(group=_GROUP_PRE_BEAMSTOP_CHECK, timeout=10)

    yield from _unload_sample_if_present(
        devices.robot,
        devices.gonio,
        devices.aperture_scatterguard,
        devices.lower_gonio,
    )

    feature_flags = get_hyperion_feature_settings()
    if feature_flags.BEAMSTOP_DIODE_CHECK:
        beamline_parameters = get_beamline_parameters(BL)
        detector_min_z = feature_flags.DETECTOR_DISTANCE_LIMIT_MIN_MM
        detector_max_z = feature_flags.DETECTOR_DISTANCE_LIMIT_MAX_MM
        yield from move_beamstop_in_and_verify_using_diode(
            devices, beamline_parameters, detector_min_z, detector_max_z
        )
    else:
        yield from bps.abs_set(
            devices.beamstop.selected_pos, BeamstopPositions.DATA_COLLECTION, wait=True
        )

    yield from bps.abs_set(
        devices.aperture_scatterguard.selected_aperture,
        ApertureValue.SMALL,
        group=_GROUP_POST_BEAMSTOP_CHECK,
    )

    yield from bps.abs_set(
        devices.cryojet.coarse, CryoInOut.IN, group=_GROUP_POST_BEAMSTOP_CHECK
    )
    yield from bps.abs_set(
        devices.cryojet.fine, CryoInOut.IN, group=_GROUP_POST_BEAMSTOP_CHECK
    )

    yield from bps.abs_set(
        devices.oav.zoom_controller, "1.0x", group=_GROUP_POST_BEAMSTOP_CHECK
    )

    yield from bps.wait(_GROUP_POST_BEAMSTOP_CHECK, timeout=10)


def _check_cryostream(devices: UDCDefaultDevices):
    commissioning_mode = yield from bps.rd(devices.baton.commissioning)
    cryo_mode = yield from bps.rd(devices.robot.cryomode_rbv)
    if commissioning_mode and cryo_mode == BartRobot.CRYO_MODE_WARM:
        LOGGER.warning("Ignoring cryostream status in commissioning mode")
    else:
        cryostream_temp = yield from bps.rd(devices.cryostream.temp)
        cryostream_pressure = yield from bps.rd(devices.cryostream.back_pressure)
        if cryostream_temp > CONST.HARDWARE.MAX_CRYO_TEMP_K:
            raise CryoStreamError(
                "Cryostream temperature is too high, not starting UDC"
            )
        if cryostream_pressure > CONST.HARDWARE.MAX_CRYO_PRESSURE_BAR:
            raise CryoStreamError(
                "Cryostream back pressure is too high, not starting UDC"
            )


def _verify_correct_cryostream_selected(
    cryostream_gantry: CryoStreamGantry,
) -> MsgGenerator:
    cryostream_selection = yield from bps.rd(cryostream_gantry.cryostream_selector)
    cryostream_selected = yield from bps.rd(cryostream_gantry.cryostream_selected)
    if cryostream_selection != CryoStreamSelection.CRYOJET or cryostream_selected != 1:
        raise CryoStreamError(
            f"Cryostream is not selected for use, control PV selection = {cryostream_selection}, "
            f"current status {cryostream_selected}"
        )


def _unload_sample_if_present(
    robot: BartRobot,
    smargon: Smargon,
    aperture_scatterguard: ApertureScatterguard,
    lower_gonio: XYZStage,
):
    pin_mounted = yield from bps.rd(robot.gonio_pin_sensor)

    if pin_mounted != PinMounted.NO_PIN_MOUNTED:
        LOGGER.info("Pin detected, unloading sample...")
        yield from robot_unload(
            robot, smargon, aperture_scatterguard, lower_gonio, None
        )



from bluesky.utils import MsgGenerator
from dodal.beamlines.i24 import CommissioningJungfrau
from dodal.common import inject
from ophyd_async.core import (
    WatchableAsyncStatus,
)
from ophyd_async.fastcs.jungfrau import (
    GainMode,
    create_jungfrau_internal_triggering_info,
)
from pydantic import PositiveInt

from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    fly_jungfrau,
)


def do_internal_acquisition(
    exp_time_s: float,
    gain_mode: GainMode,
    total_frames: PositiveInt = 1,
    jungfrau: CommissioningJungfrau = inject("jungfrau"),
    path_of_output_file: str | None = None,
    wait: bool = False,
) -> MsgGenerator[WatchableAsyncStatus]:
    """
    Kickoff internal triggering on the Jungfrau, and optionally wait for completion. Frames
    per trigger will trigger as rapidly as possible according to the Jungfrau deadtime.

    Any plan using this stub MUST stage the Jungfrau with the stage_decorator and open a run,
    ideally using the run_decorator.

    Args:
        exp_time_s: Length of detector exposure for each frame.
        gain_mode: Which gain mode to put the Jungfrau into before starting the acquisition.
        total_frames: Number of frames taken after being internally triggered.
        period_between_frames_s: Time between each detector frame, including deadtime. Not needed if frames_per_triggers is 1.
        jungfrau: Jungfrau device
        path_of_output_file: Absolute path of the detector file output, including file name. If None, then use the PathProvider
            set during jungfrau device instantiation
        wait: Optionally block until data collection is complete.
    """

    trigger_info = create_jungfrau_internal_triggering_info(total_frames, exp_time_s)
    status = yield from fly_jungfrau(jungfrau, trigger_info, gain_mode, wait=wait)
    return status



from daq_config_server.models.feature_settings.hyperion_feature_settings import (
    HyperionFeatureSettings,
)
from dodal.beamlines.i03 import DAQ_CONFIGURATION_PATH
from dodal.common.beamlines.config_client import get_config_client

GDA_DOMAIN_PROPERTIES_PATH = DAQ_CONFIGURATION_PATH + "/domain/domain.properties"


def get_hyperion_feature_settings() -> HyperionFeatureSettings:
    config_client = get_config_client("i03")
    return config_client.get_file_contents(
        GDA_DOMAIN_PROPERTIES_PATH,
        desired_return_type=HyperionFeatureSettings,
    )



from contextlib import nullcontext
from unittest.mock import ANY, MagicMock, patch

import pytest
from bluesky import Msg
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from daq_config_server.models.feature_settings.hyperion_feature_settings import (
    HyperionFeatureSettings,
)
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.collimation_table import CollimationTable
from dodal.devices.cryostream import (
    CryoStreamGantry,
    CryoStreamSelection,
    OxfordCryoJet,
    OxfordCryoStream,
)
from dodal.devices.cryostream import InOut as CryoInOut
from dodal.devices.fluorescence_detector_motion import FluorescenceDetector
from dodal.devices.fluorescence_detector_motion import InOut as FlouInOut
from dodal.devices.hutch_shutter import (
    HutchInterlock,
    InterlockedHutchShutter,
    ShutterDemand,
)
from dodal.devices.motors import XYZStage
from dodal.devices.mx_phase1.beamstop import BeamstopPositions
from dodal.devices.robot import BartRobot, PinMounted
from dodal.devices.scintillator import InOut, Scintillator
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutterState
from ophyd_async.core import Device, Signal, init_devices, set_mock_value
from ophyd_async.epics.motor import Motor

from mx_bluesky.hyperion.experiment_plans.udc_default_state import (
    CryoStreamError,
    UDCDefaultDevices,
    move_to_udc_default_state,
)
from mx_bluesky.hyperion.parameters.constants import CONST


@pytest.fixture
async def cryostream_gantry(sim_run_engine: RunEngineSimulator):
    async with init_devices(mock=True):
        cryostream_gantry = CryoStreamGantry("")

    set_mock_value(cryostream_gantry.cryostream_selector, CryoStreamSelection.CRYOJET)
    set_mock_value(cryostream_gantry.cryostream_selected, 1)
    sim_run_engine.add_read_handler_for(
        cryostream_gantry.cryostream_selector, CryoStreamSelection.CRYOJET
    )
    sim_run_engine.add_read_handler_for(cryostream_gantry.cryostream_selected, 1)
    yield cryostream_gantry


@pytest.fixture
async def default_devices(
    beamstop_check_devices,
    cryostream_gantry,
    robot,
    smargon,
    oav,
    sim_run_engine,
    run_engine,
):
    async with init_devices(mock=True):
        cryostream = OxfordCryoStream("")
        cryojet = OxfordCryoJet("")
        fluo = FluorescenceDetector("")
        hutch_shutter = InterlockedHutchShutter("", interlock=HutchInterlock(""))
        scintillator = Scintillator("", MagicMock(), MagicMock(), name="scin")
        collimation_table = CollimationTable("")
        lower_gonio = XYZStage("")

    with patch("dodal.devices.hutch_shutter.TEST_MODE", True):
        devices = UDCDefaultDevices(
            collimation_table=collimation_table,
            cryostream=cryostream,
            cryojet=cryojet,
            cryostream_gantry=cryostream_gantry,
            fluorescence_det_motion=fluo,
            hutch_shutter=hutch_shutter,
            lower_gonio=lower_gonio,
            robot=robot,
            scintillator=scintillator,
            gonio=smargon,
            oav=oav,
            **beamstop_check_devices.__dict__,
        )
        sim_run_engine.add_read_handler_for(
            devices.robot.gonio_pin_sensor, PinMounted.NO_PIN_MOUNTED
        )

        yield devices


@pytest.fixture
def feature_flags_with_beamstop_diode_check():
    with patch(
        "mx_bluesky.hyperion.experiment_plans.udc_default_state.get_hyperion_feature_settings"
    ) as mock_get_feature_settings:
        mock_get_feature_settings.return_value = HyperionFeatureSettings(
            BEAMSTOP_DIODE_CHECK=True,
        )
        yield mock_get_feature_settings.return_value


async def test_given_cryostream_temp_is_too_high_then_exception_raised(
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    sim_run_engine.add_read_handler_for(
        default_devices.cryostream.temp,
        CONST.HARDWARE.MAX_CRYO_TEMP_K + 10,
    )
    with pytest.raises(CryoStreamError, match="temperature is too high"):
        sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))


async def test_given_cryostream_pressure_is_too_high_then_exception_raised(
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    sim_run_engine.add_read_handler_for(
        default_devices.cryostream.back_pressure,
        CONST.HARDWARE.MAX_CRYO_PRESSURE_BAR + 10,
    )
    with pytest.raises(CryoStreamError, match="pressure is too high"):
        sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))


async def test_scintillator_is_moved_out_before_aperture_scatterguard_moved_in(
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "set"
            and msg.obj.name == "scin-selected_pos"
            and msg.args[0] == InOut.OUT
        ),
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "set"
            and msg.obj.name == "aperture_scatterguard-selected_aperture"
            and msg.args[0] == ApertureValue.SMALL
        ),
    )


def test_udc_default_state_runs_in_real_run_engine(
    run_engine: RunEngine, default_devices: UDCDefaultDevices
):
    set_mock_value(default_devices.cryostream.temp, 100)
    set_mock_value(default_devices.cryostream.back_pressure, 0.01)
    default_devices.scintillator._aperture_scatterguard().selected_aperture.get_value = MagicMock(
        return_value=ApertureValue.PARKED
    )

    run_engine(move_to_udc_default_state(default_devices))


@patch(
    "mx_bluesky.hyperion.experiment_plans.udc_default_state._unload_sample_if_present",
    MagicMock(return_value=iter([Msg("robot_unload")])),
)
def test_beamstop_moved_to_data_collection_if_diode_check_not_enabled(
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    pre_beamstop_group = "pre_beamstop_check"
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == pre_beamstop_group,
    )
    assert msgs[1].command == "robot_unload"
    assert (
        msgs[2].command == "set"
        and msgs[2].obj is default_devices.beamstop.selected_pos
        and msgs[2].args[0] == BeamstopPositions.DATA_COLLECTION
    )
    assert (
        msgs[3].command == "wait" and msgs[3].kwargs["group"] == msgs[2].kwargs["group"]
    )


@pytest.mark.parametrize(
    "min_z, max_z",
    [
        [250, 800],
        [300, 650],
    ],
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.udc_default_state.move_beamstop_in_and_verify_using_diode",
    return_value=iter([Msg("move_beamstop_in")]),
)
def test_beamstop_check_is_called_with_detector_distances_from_config_server(
    mock_move_beamstop_in: MagicMock,
    default_devices: UDCDefaultDevices,
    sim_run_engine: RunEngineSimulator,
    min_z: float,
    max_z: float,
):
    with patch(
        "mx_bluesky.hyperion.experiment_plans.udc_default_state.get_hyperion_feature_settings"
    ) as mock_get_feature_settings:
        mock_get_feature_settings.return_value = HyperionFeatureSettings(
            BEAMSTOP_DIODE_CHECK=True,
            DETECTOR_DISTANCE_LIMIT_MAX_MM=max_z,
            DETECTOR_DISTANCE_LIMIT_MIN_MM=min_z,
        )
        sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))
    mock_move_beamstop_in.assert_called_once_with(ANY, ANY, min_z, max_z)


@patch(
    "mx_bluesky.hyperion.experiment_plans.udc_default_state.move_beamstop_in_and_verify_using_diode",
    MagicMock(return_value=iter([Msg("move_beamstop_in")])),
)
def test_udc_pre_and_post_groups_contains_expected_items_and_are_waited_on_before_and_after_beamstop_check(
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
    feature_flags_with_beamstop_diode_check: HyperionFeatureSettings,
):
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))

    pre_beamstop_group = "pre_beamstop_check"
    post_beamstop_group = "post_beamstop_check"

    def assert_expected_set(signal: Signal | Motor | Device, value, group):
        return assert_message_and_return_remaining(
            msgs,
            lambda msg: (
                msg.command == "set"
                and msg.obj.name == signal.name
                and msg.args[0] == value
                and msg.kwargs["group"] == group
            ),
        )

    msgs = assert_expected_set(
        default_devices.fluorescence_det_motion.pos, FlouInOut.OUT, pre_beamstop_group
    )
    coll = default_devices.collimation_table
    for device in [
        coll.inboard_y,
        coll.outboard_y,
        coll.upstream_y,
        coll.upstream_x,
        coll.downstream_x,
    ]:
        msgs = assert_expected_set(device, 0, pre_beamstop_group)

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == pre_beamstop_group,
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "move_beamstop_in"
    )

    msgs = assert_expected_set(
        default_devices.aperture_scatterguard.selected_aperture,
        ApertureValue.SMALL,
        post_beamstop_group,
    )

    msgs = assert_expected_set(
        default_devices.cryojet.coarse, CryoInOut.IN, post_beamstop_group
    )
    msgs = assert_expected_set(
        default_devices.cryojet.fine, CryoInOut.IN, post_beamstop_group
    )

    msgs = assert_expected_set(
        default_devices.oav.zoom_controller, "1.0x", post_beamstop_group
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "wait" and msg.kwargs["group"] == post_beamstop_group
        ),
    )


@pytest.mark.parametrize(
    "expected_raise, cryostream_selection, cryostream_selected",
    [
        [nullcontext(), CryoStreamSelection.CRYOJET, 1],
        [pytest.raises(CryoStreamError), CryoStreamSelection.HC1, 1],
        [pytest.raises(CryoStreamError), CryoStreamSelection.CRYOJET, 0],
    ],
)
def test_udc_default_state_checks_cryostream_selection(
    run_engine: RunEngine,
    default_devices,
    expected_raise,
    cryostream_selection: CryoStreamSelection,
    cryostream_selected: int,
):
    default_devices.scintillator._aperture_scatterguard().selected_aperture.get_value = MagicMock(
        return_value=ApertureValue.PARKED
    )
    set_mock_value(
        default_devices.cryostream_gantry.cryostream_selector, cryostream_selection
    )
    set_mock_value(
        default_devices.cryostream_gantry.cryostream_selected, cryostream_selected
    )

    with expected_raise:
        run_engine(move_to_udc_default_state(default_devices))


@patch(
    "mx_bluesky.hyperion.experiment_plans.udc_default_state._verify_correct_cryostream_selected",
    MagicMock(return_value=iter([Msg("cryostream_gantry_check")])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.udc_default_state._check_cryostream",
    MagicMock(return_value=iter([Msg("cryostream_check")])),
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.udc_default_state.robot_unload",
    MagicMock(return_value=iter([Msg("unload")])),
)
def test_udc_default_state_unloads_if_sample_present(
    default_devices, sim_run_engine, beamline_parameters
):
    sim_run_engine.add_read_handler_for(
        default_devices.robot.gonio_pin_sensor, PinMounted.PIN_MOUNTED
    )
    with patch(
        "mx_bluesky.hyperion.experiment_plans.udc_default_state.get_beamline_parameters",
        return_value=beamline_parameters,
    ):
        msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "cryostream_gantry_check"
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "cryostream_check"
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "set"
            and msg.obj == default_devices.scintillator.selected_pos
            and msg.args[0] == InOut.OUT
        ),
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "unload"
    )


def test_udc_default_state_no_unload_if_no_sample_present(
    default_devices, sim_run_engine, beamline_parameters
):
    sim_run_engine.add_read_handler_for(
        default_devices.robot.gonio_pin_sensor, PinMounted.NO_PIN_MOUNTED
    )
    with patch(
        "mx_bluesky.hyperion.experiment_plans.udc_default_state.get_beamline_parameters",
        return_value=beamline_parameters,
    ):
        msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))

    assert not [m for m in msgs if m.command == "unload"]


def test_default_state_closes_sample_shutter_before_open_hutch_shutter(
    sim_run_engine: RunEngineSimulator, default_devices: UDCDefaultDevices
):
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "set"
            and msg.obj is default_devices.sample_shutter
            and msg.args[0] == ZebraShutterState.CLOSE
        ),
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "wait" and msg.kwargs["group"] == msgs[0].kwargs["group"]
        ),
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: (
            msg.command == "set"
            and msg.obj is default_devices.hutch_shutter
            and msg.args[0] == ShutterDemand.OPEN
            and msg.kwargs["group"] == "pre_beamstop_check"
        ),
    )


def test_default_state_hutch_shutter_open_is_skipped_if_commissioning_mode_enabled(
    sim_run_engine: RunEngineSimulator, default_devices: UDCDefaultDevices
):
    sim_run_engine.add_read_handler_for(default_devices.baton.commissioning, True)
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))
    open_shutter_msgs = [
        msg
        for msg in msgs
        if msg.command == "set"
        and msg.obj is default_devices.hutch_shutter
        and msg.args[0] == ShutterDemand.OPEN
    ]
    assert not open_shutter_msgs


@pytest.mark.parametrize(
    "cryo_mode, commissioning_mode, expected_exception",
    [
        [BartRobot.CRYO_MODE_WARM, True, None],
        [BartRobot.CRYO_MODE_CRYO, True, CryoStreamError],
        [BartRobot.CRYO_MODE_WARM, False, CryoStreamError],
        [BartRobot.CRYO_MODE_CRYO, False, CryoStreamError],
    ],
)
def test_default_state_cryo_check_skipped_in_commissioning_mode_if_cryo_mode(
    run_engine: RunEngine,
    default_devices: UDCDefaultDevices,
    cryo_mode: float,
    commissioning_mode: bool,
    expected_exception: type | None,
):
    set_mock_value(default_devices.robot.cryomode_rbv, cryo_mode)
    set_mock_value(default_devices.baton.commissioning, commissioning_mode)
    set_mock_value(
        default_devices.cryostream.temp,
        CONST.HARDWARE.MAX_CRYO_TEMP_K + 1,
    )
    set_mock_value(
        default_devices.cryostream.back_pressure,
        CONST.HARDWARE.MAX_CRYO_PRESSURE_BAR + 0.1,
    )
    with pytest.raises(expected_exception) if expected_exception else nullcontext():
        run_engine(move_to_udc_default_state(default_devices))



import datetime
import os


def get_ispyb_config() -> str:
    ispyb_config = os.environ.get("ISPYB_CONFIG_PATH")
    assert ispyb_config, "ISPYB_CONFIG_PATH must be set"
    return ispyb_config


def get_current_time_string():
    now = datetime.datetime.now()
    return now.strftime("%Y-%m-%d %H:%M:%S")



from __future__ import annotations

from pathlib import Path
from typing import Protocol, TypeVar

from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.devices.backlight import InOut
from dodal.devices.eiger import EigerDetector
from dodal.devices.oav.oav_parameters import OAVParameters

from mx_bluesky.common.device_setup_plans.manipulate_sample import (
    move_aperture_if_required,
)
from mx_bluesky.common.device_setup_plans.utils import (
    start_preparing_data_collection_then_do_plan,
)
from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    common_flyscan_xray_centre,
)
from mx_bluesky.common.experiment_plans.oav_grid_detection_plan import (
    OavGridDetectionComposite,
    grid_detection_plan,
)
from mx_bluesky.common.experiment_plans.oav_snapshot_plan import (
    setup_beamline_for_oav,
)
from mx_bluesky.common.external_interaction.callbacks.common.grid_detection_callback import (
    GridDetectionCallback,
    GridParamUpdate,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    ispyb_activation_wrapper,
)
from mx_bluesky.common.parameters.constants import (
    OavConstants,
    PlanGroupCheckpointConstants,
)
from mx_bluesky.common.parameters.device_composites import (
    FlyScanEssentialDevices,
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.common.parameters.gridscan import GridCommon, SpecifiedThreeDGridScan
from mx_bluesky.common.utils.log import LOGGER

TFlyScanEssentialDevices = TypeVar(
    "TFlyScanEssentialDevices", bound=FlyScanEssentialDevices, contravariant=True
)
TSpecifiedThreeDGridScan = TypeVar(
    "TSpecifiedThreeDGridScan", bound=SpecifiedThreeDGridScan, contravariant=True
)


def grid_detect_then_xray_centre(
    composite: GridDetectThenXRayCentreComposite,
    parameters: GridCommon,
    xrc_params_type: type[SpecifiedThreeDGridScan],
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
    oav_config: str = OavConstants.OAV_CONFIG_JSON,
) -> MsgGenerator:
    """
    A plan which combines the collection of snapshots from the OAV and the determination
    of the grid dimensions to use for the following grid scan.
    """

    eiger: EigerDetector = composite.eiger

    eiger.set_detector_parameters(parameters.detector_params)

    oav_params = OAVParameters("xrayCentring", oav_config)

    def plan_to_perform():
        yield from ispyb_activation_wrapper(
            detect_grid_and_do_gridscan(
                composite,
                parameters,
                oav_params,
                xrc_params_type,
                construct_beamline_specific,
            ),
            parameters,
        )

    yield from start_preparing_data_collection_then_do_plan(
        composite.beamstop,
        eiger,
        composite.detector_motion,
        parameters.detector_params.detector_distance,
        plan_to_perform(),
        group=PlanGroupCheckpointConstants.GRID_READY_FOR_DC,
    )


# This function should be private but is currently called by Hyperion, see https://github.com/DiamondLightSource/mx-bluesky/issues/1148
def detect_grid_and_do_gridscan(
    composite: GridDetectThenXRayCentreComposite,
    parameters: GridCommon,
    oav_params: OAVParameters,
    xrc_params_type: type[SpecifiedThreeDGridScan],
    construct_beamline_specific: ConstructBeamlineSpecificFeatures,
):
    snapshot_template = f"{parameters.detector_params.prefix}_{parameters.detector_params.run_number}_{{angle}}"

    grid_params_callback = GridDetectionCallback()

    yield from setup_beamline_for_oav(
        composite.gonio,
        composite.backlight,
        composite.aperture_scatterguard,
        wait=True,
    )

    @bpp.subs_decorator([grid_params_callback])
    def run_grid_detection_plan(
        oav_params,
        snapshot_template,
        snapshot_dir: Path,
    ):
        grid_detect_composite = OavGridDetectionComposite(
            backlight=composite.backlight,
            oav=composite.oav,
            gonio=composite.gonio,
            pin_tip_detection=composite.pin_tip_detection,
        )

        yield from grid_detection_plan(
            grid_detect_composite,
            oav_params,
            snapshot_template,
            str(snapshot_dir),
            parameters.grid_width_um,
            parameters.box_size_um,
        )

    if parameters.selected_aperture:
        # Start moving the aperture/scatterguard into position without moving it in
        yield from bps.prepare(
            composite.aperture_scatterguard,
            parameters.selected_aperture,
            group=PlanGroupCheckpointConstants.PREPARE_APERTURE,
        )

    yield from run_grid_detection_plan(
        oav_params,
        snapshot_template,
        parameters.snapshot_directory,
    )

    yield from bps.abs_set(
        composite.backlight,
        InOut.OUT,
        group=PlanGroupCheckpointConstants.GRID_READY_FOR_DC,
    )

    yield from bps.wait(PlanGroupCheckpointConstants.PREPARE_APERTURE)
    yield from move_aperture_if_required(
        composite.aperture_scatterguard,
        parameters.selected_aperture,
        group=PlanGroupCheckpointConstants.GRID_READY_FOR_DC,
    )
    xrc_params = create_parameters_for_flyscan_xray_centre(
        parameters, grid_params_callback.get_grid_parameters(), xrc_params_type
    )
    parameters.set_specified_grid_params(xrc_params)
    beamline_specific = construct_beamline_specific(composite, xrc_params)

    yield from common_flyscan_xray_centre(composite, xrc_params, beamline_specific)


class ConstructBeamlineSpecificFeatures(
    Protocol[TFlyScanEssentialDevices, TSpecifiedThreeDGridScan]
):
    def __call__(
        self,
        xrc_composite: TFlyScanEssentialDevices,
        xrc_parameters: TSpecifiedThreeDGridScan,
    ) -> BeamlineSpecificFGSFeatures: ...


def create_parameters_for_flyscan_xray_centre(
    parameters: GridCommon,
    grid_parameters: GridParamUpdate,
    xrc_params_type: type[SpecifiedThreeDGridScan],
) -> SpecifiedThreeDGridScan:
    params_json = parameters.model_dump()
    params_json.update(grid_parameters)
    flyscan_xray_centre_parameters = xrc_params_type(**params_json)
    LOGGER.info(f"Parameters for FGS: {flyscan_xray_centre_parameters}")
    return flyscan_xray_centre_parameters



from __future__ import annotations

from functools import partial
from threading import Event
from time import sleep
from unittest.mock import MagicMock, Mock, call, patch

import pytest
from bluesky.callbacks import CallbackBase
from bluesky_stomp.models import Broker
from dodal.log import LOGGER as DODAL_LOGGER

from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER, NEXUS_LOGGER
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    PING_TIMEOUT_S,
    main,
    ping_watchdog_while_alive,
    run_watchdog,
    setup_callbacks,
    setup_logging,
)
from mx_bluesky.hyperion.parameters.cli import CallbackArgs
from mx_bluesky.hyperion.parameters.constants import HyperionConstants


@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.run_watchdog")
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.parse_callback_args",
    return_value=CallbackArgs(True, HyperionConstants.SUPERVISOR_PORT),
)
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.setup_callbacks")
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.setup_logging")
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.set_alerting_service"
)
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.RemoteDispatcher")
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.Proxy")
def test_main_function(
    mock_proxy: MagicMock,
    mock_dispatcher: MagicMock,
    setup_alerting: MagicMock,
    setup_logging: MagicMock,
    setup_callbacks: MagicMock,
    parse_callback_args: MagicMock,
    mock_run_watchdog: MagicMock,
):
    proxy_started = Event()
    dispatcher_started = Event()
    watchdog_started = Event()
    mock_proxy.return_value.start.side_effect = proxy_started.set
    mock_dispatcher.return_value.start.side_effect = dispatcher_started.set
    mock_run_watchdog.side_effect = lambda _: watchdog_started.set() or None

    main()

    proxy_started.wait(0.5)
    dispatcher_started.wait(0.5)
    mock_run_watchdog.wait(0.5)
    setup_logging.assert_called()
    setup_callbacks.assert_called()
    setup_alerting.assert_called_once()
    mock_run_watchdog.assert_called_once()
    assert isinstance(setup_alerting.mock_calls[0].args[0], LoggingAlertService)


def test_setup_callbacks():
    current_number_of_callbacks = 8
    cbs = setup_callbacks()
    assert len(cbs) == current_number_of_callbacks
    assert len(set(cbs)) == current_number_of_callbacks


@pytest.mark.skip_log_setup
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.parse_callback_args",
    return_value=CallbackArgs(True, HyperionConstants.SUPERVISOR_PORT),
)
def test_setup_logging(parse_callback_cli_args):
    assert DODAL_LOGGER.parent != ISPYB_ZOCALO_CALLBACK_LOGGER
    assert len(ISPYB_ZOCALO_CALLBACK_LOGGER.handlers) == 0
    assert len(NEXUS_LOGGER.handlers) == 0
    setup_logging(parse_callback_cli_args())
    assert len(ISPYB_ZOCALO_CALLBACK_LOGGER.handlers) == 4
    assert len(NEXUS_LOGGER.handlers) == 4
    assert DODAL_LOGGER.parent == ISPYB_ZOCALO_CALLBACK_LOGGER
    setup_logging(parse_callback_cli_args())
    assert len(ISPYB_ZOCALO_CALLBACK_LOGGER.handlers) == 4
    assert len(NEXUS_LOGGER.handlers) == 4


@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.sleep")
def test_wait_for_threads_forever_calls_time_sleep(mock_sleep: MagicMock):
    thread_that_stops_after_one_call = MagicMock()
    thread_that_stops_after_one_call.is_alive.side_effect = [True, False]

    mock_context_manager = MagicMock()
    mock_context_manager.is_alive.side_effect = [True, False]

    ping_watchdog_while_alive(thread_that_stops_after_one_call, mock_context_manager)
    assert mock_sleep.call_count == 1


@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.RemoteDispatcher")
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.Proxy")
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.request")
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.LIVENESS_POLL_SECONDS",
    0.1,
)
def test_launching_external_callbacks_pings_regularly(
    mock_request: MagicMock,
    mock_proxy: MagicMock,
    mock_dispatcher: MagicMock,
):
    mock_proxy.return_value.start.side_effect = partial(sleep, 0.1)
    mock_dispatcher.return_value.start.side_effect = partial(sleep, 0.1)
    mock_request.urlopen.return_value.__enter__.return_value.status = 200
    mock_request.urlopen.return_value.__exit__.side_effect = RuntimeError(
        "Exit this thread"
    )

    with pytest.raises(RuntimeError, match="Exit this thread"):
        run_watchdog(5005)

    mock_request.urlopen.assert_called_with(
        "http://localhost:5005/callbackPing", timeout=PING_TIMEOUT_S
    )


@patch("sys.argv", new=["hyperion-callbacks", "--watchdog-port", "1234"])
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.HyperionCallbackRunner"
)
def test_launch_with_watchdog_port_arg_applies_port(mock_callback_runner: MagicMock):
    main(dev_mode=True)
    mock_callback_runner.assert_called_once()
    callback_args = mock_callback_runner.mock_calls[0].args[0]
    assert callback_args.dev_mode
    assert callback_args.watchdog_port == 1234


@patch(
    "sys.argv",
    new=[
        "hyperion-callbacks",
        "--watchdog-port",
        "1234",
        "--stomp-config",
        "tests/test_data/stomp_callback_test_config.yaml",
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.StompDispatcher")
@patch("mx_bluesky.hyperion.external_interaction.callbacks.__main__.StompClient")
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.setup_callbacks",
    return_value=[Mock(spec=CallbackBase)],
)
@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.__main__.LIVENESS_POLL_SECONDS",
    0.1,
)
def test_launch_with_stomp_launches_stomp_backend(
    mock_setup_callbacks: MagicMock,
    mock_client_cls: MagicMock,
    mock_dispatcher_cls: MagicMock,
):
    stomp_client = mock_client_cls.for_broker.return_value
    dispatcher = mock_dispatcher_cls.return_value
    stomp_client.is_connected.side_effect = [True, False]

    parent = MagicMock()
    parent.attach_mock(stomp_client, "stomp_client")
    parent.attach_mock(dispatcher, "dispatcher")
    main(dev_mode=True)

    mock_client_cls.for_broker.assert_called_once_with(
        broker=Broker(host="localhost", port=61613, auth=None)
    )
    mock_dispatcher_cls.assert_called_once_with(stomp_client)
    parent.assert_has_calls(
        [
            call.dispatcher.subscribe(mock_setup_callbacks.return_value[0]),
            call.dispatcher.__enter__(),
            call.stomp_client.is_connected(),
            call.stomp_client.is_connected(),
            call.dispatcher.__exit__(None, None, None),
        ]
    )



from collections.abc import Callable
from typing import cast

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common.watcher_utils import log_on_percentage_complete
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from ophyd_async.core import (
    TriggerInfo,
    WatchableAsyncStatus,
)
from ophyd_async.fastcs.jungfrau import GainMode

from mx_bluesky.common.utils.log import LOGGER

JF_COMPLETE_GROUP = "JF complete"


def fly_jungfrau(
    jungfrau: CommissioningJungfrau,
    trigger_info: TriggerInfo,
    gain_mode: GainMode,
    wait: bool = False,
    log_on_percentage_prefix="Jungfrau data collection triggers received",
    read_hardware_after_prepare_plan: Callable[..., MsgGenerator]
    | None = None,  # Param needs refactor: https://github.com/DiamondLightSource/mx-bluesky/issues/819
) -> MsgGenerator[WatchableAsyncStatus]:
    """Stage, prepare, and kickoff Jungfrau with a configured TriggerInfo. Optionally wait
    for completion.

    Any plan using this stub MUST stage the Jungfrau with the stage_decorator and open a run,
    ideally using the run_decorator.

    Args:
    jungfrau: Jungfrau device.
    trigger_info: TriggerInfo which should be acquired using jungfrau util functions.
    gain_mode: Which gain mode to put the Jungfrau into before starting the acquisition.
    wait: Optionally block until data collection is complete.
    log_on_percentage_prefix: String that will be appended to the "percentage completion" logging message.
    read_hardware_after_prepare_plan: Optionally add a plan which will be ran in between preparing the jungfrau and starting
    acquisition. This is useful for reading devices after they have been prepared, especially since the file writing path
    is calculated during prepare.
    """

    LOGGER.info(f"Setting Jungfrau to gain mode {gain_mode}")
    yield from bps.mv(jungfrau.drv.gain_mode, gain_mode)
    LOGGER.info("Preparing detector...")
    yield from bps.prepare(jungfrau, trigger_info, wait=True)
    LOGGER.info("Detector prepared")
    if read_hardware_after_prepare_plan:
        yield from read_hardware_after_prepare_plan()
    yield from bps.kickoff(jungfrau, wait=True)
    LOGGER.info("Waiting for acquisition to complete...")
    status = yield from bps.complete(jungfrau, group=JF_COMPLETE_GROUP)

    # StandardDetector.complete converts regular status to watchable status,
    # but bluesky plan stubs can't see this currently
    status = cast(WatchableAsyncStatus, status)
    log_on_percentage_complete(status, log_on_percentage_prefix, 10)
    if wait:
        yield from bps.wait(JF_COMPLETE_GROUP)
    return status



import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.preprocessors import contingency_decorator, run_decorator, subs_decorator
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.beamlines.i04.constants import RedisConstants
from dodal.devices.beamlines.i04.murko_results import MurkoResultsDevice
from dodal.devices.oav.oav_to_redis_forwarder import OAVToRedisForwarder, Source
from dodal.devices.robot import BartRobot
from dodal.devices.smargon import Smargon
from dodal.devices.thawer import OnOff, Thawer
from dodal.log import LOGGER

from mx_bluesky.beamlines.i04.callbacks.murko_callback import MurkoCallback


def thaw(
    time_to_thaw: float,
    rotation: float = 360,
    thawer: Thawer = inject("thawer"),
    smargon: Smargon = inject("gonio"),
) -> MsgGenerator:
    """Turns on the thawer and rotates the sample by {rotation} degrees to thaw it, then
    rotates {rotation} degrees back and turns the thawer off. The speed of the goniometer
    is set such that the process takes whole process will take {time_to_thaw} time.

    Args:
        time_to_thaw (float): Time to thaw for, in seconds.
        rotation (float): How much to rotate by whilst thawing, in degrees.
        thawer (Thawer): The thawing device.
        smargon (Smargon): The smargon used to rotate.
    """
    initial_velocity = yield from bps.rd(smargon.omega.velocity)
    new_velocity = abs(rotation / time_to_thaw) * 2.0

    def do_thaw():
        yield from bps.abs_set(smargon.omega.velocity, new_velocity, wait=True)
        yield from bps.abs_set(thawer, OnOff.ON, wait=True)
        yield from bps.rel_set(smargon.omega, rotation, wait=True)
        yield from bps.rel_set(smargon.omega, -rotation, wait=True)

    def cleanup():
        yield from bps.abs_set(smargon.omega.velocity, initial_velocity, wait=True)
        yield from bps.abs_set(thawer, OnOff.OFF, wait=True)

    yield from bpp.contingency_wrapper(
        do_thaw(),
        final_plan=cleanup,
    )


def thaw_and_murko_centre(
    time_to_thaw: float,
    rotation: float = 360,
    robot: BartRobot = inject("robot"),
    thawer: Thawer = inject("thawer"),
    smargon: Smargon = inject("gonio"),
    murko_results: MurkoResultsDevice = inject("murko_results"),
    oav_to_redis_forwarder: OAVToRedisForwarder = inject("oav_to_redis_forwarder"),
) -> MsgGenerator:
    """Thaws the sample and centres it using murko by:
        1. Turns on the thawer
        2. Rotates the sample by {rotation} degrees, whilst this is happening images from
        the full screen OAV are being fed to murko
        3. After the rotation has completed moves to the average centre returned by murko
        from these images
        4. Rotate {rotation} degrees back to the start, whilst this is happening images
        from the small ROI of the OAV are being fed to murko
        5. Turns off the thawer

    The speed of the goniometer is set so that all of the above takes about {time_to_thaw}
    seconds to complete.

    Args:
        time_to_thaw (float): Time to thaw for, in seconds.
        rotation (float, optional): How much to rotate by whilst thawing, in degrees.
                                    Defaults to 360.
        ... devices: These are the specific ophyd-devices used for the plan, the
                     defaults are always correct
    """
    murko_results_group = "get_results"

    sample_id = yield from bps.rd(robot.sample_id)
    sample_id = int(sample_id)

    oav_fs = oav_to_redis_forwarder.sources[Source.FULL_SCREEN].oav_ref()

    initial_zoom_level = yield from bps.rd(oav_fs.zoom_controller.level)
    initial_velocity = yield from bps.rd(smargon.omega.velocity)
    new_velocity = abs(rotation / time_to_thaw) * 2.0

    murko_callback = MurkoCallback(
        RedisConstants.REDIS_HOST,
        RedisConstants.REDIS_PASSWORD,
        RedisConstants.MURKO_REDIS_DB,
    )

    def cleanup():
        yield from bps.mv(oav_fs.zoom_controller.level, initial_zoom_level)
        yield from bps.abs_set(smargon.omega.velocity, initial_velocity, wait=True)
        yield from bps.abs_set(thawer, OnOff.OFF, wait=True)

    def centre_from_murko():
        yield from bps.wait(murko_results_group)

        x_predict = yield from bps.rd(murko_results.x_mm)
        y_predict = yield from bps.rd(murko_results.y_mm)
        z_predict = yield from bps.rd(murko_results.z_mm)

        LOGGER.info(f"Got results: {x_predict, y_predict, z_predict}")

        yield from bps.rel_set(smargon.x, x_predict)
        yield from bps.rel_set(smargon.y, y_predict)
        yield from bps.rel_set(smargon.z, z_predict)

    @subs_decorator(murko_callback)
    @contingency_decorator(final_plan=cleanup)
    def do_thaw_and_murko_centre():
        yield from bps.mv(
            murko_results.sample_id,
            str(sample_id),
            oav_to_redis_forwarder.sample_id,
            sample_id,
            oav_fs.zoom_controller.level,
            "1.0x",
        )
        yield from bps.abs_set(smargon.omega.velocity, new_velocity, wait=True)
        yield from bps.abs_set(thawer, OnOff.ON, wait=True)

        def rotate_in_one_direction_then_murko_centre(
            rotation: float, oav_mode: Source
        ):
            @run_decorator(md={"sample_id": sample_id})
            def rotate_in_one_direction_and_start_murko_and_stream_to_redis():
                yield from bps.stage(murko_results, wait=True)
                yield from bps.trigger(murko_results, group=murko_results_group)

                yield from _rotate_in_one_direction_and_stream_to_redis(
                    smargon, oav_to_redis_forwarder, oav_mode, rotation
                )

            yield from rotate_in_one_direction_and_start_murko_and_stream_to_redis()

            yield from centre_from_murko()
            yield from bps.unstage(murko_results, wait=True)

        yield from rotate_in_one_direction_then_murko_centre(
            rotation, Source.FULL_SCREEN
        )
        yield from rotate_in_one_direction_then_murko_centre(-rotation, Source.ROI)

    yield from do_thaw_and_murko_centre()


def thaw_and_stream_to_redis(
    time_to_thaw: float,
    rotation: float = 360,
    robot: BartRobot = inject("robot"),
    thawer: Thawer = inject("thawer"),
    smargon: Smargon = inject("gonio"),
    oav_to_redis_forwarder: OAVToRedisForwarder = inject("oav_to_redis_forwarder"),
) -> MsgGenerator:
    """Turns on the thawer and rotates the sample by {rotation} degrees to thaw it, then
    rotates {rotation} degrees back and turns the thawer off. The speed of the goniometer
    is set such that the process takes whole process will take {time_to_thaw} time.

    At the same time streams OAV images to redis for later processing (e.g. by murko).
    On the first rotation the images from the large ROI are streamed, on the second the
    smaller ROI is used.

    Args:
        time_to_thaw (float): Time to thaw for, in seconds.
        rotation (float, optional): How much to rotate by whilst thawing, in degrees.
                                    Defaults to 360.
        ... devices: These are the specific ophyd-devices used for the plan, the
                     defaults are always correct
    """
    sample_id = yield from bps.rd(robot.sample_id)
    sample_id = int(sample_id)

    oav_fs = oav_to_redis_forwarder.sources[Source.FULL_SCREEN].oav_ref()

    initial_zoom_level = yield from bps.rd(oav_fs.zoom_controller.level)
    initial_velocity = yield from bps.rd(smargon.omega.velocity)
    new_velocity = abs(rotation / time_to_thaw) * 2.0

    murko_callback = MurkoCallback(
        RedisConstants.REDIS_HOST,
        RedisConstants.REDIS_PASSWORD,
        RedisConstants.MURKO_REDIS_DB,
    )

    def cleanup():
        yield from bps.mv(oav_fs.zoom_controller.level, initial_zoom_level)
        yield from bps.abs_set(smargon.omega.velocity, initial_velocity, wait=True)
        yield from bps.abs_set(thawer, OnOff.OFF, wait=True)

    @subs_decorator(murko_callback)
    @contingency_decorator(final_plan=cleanup)
    def do_thaw_and_stream_to_redis():
        yield from bps.mv(
            oav_to_redis_forwarder.sample_id,
            sample_id,
            oav_fs.zoom_controller.level,
            "1.0x",
        )
        yield from bps.abs_set(smargon.omega.velocity, new_velocity, wait=True)
        yield from bps.abs_set(thawer, OnOff.ON, wait=True)

        @run_decorator(md={"sample_id": sample_id})
        def rotate_in_one_direction_and_stream_to_redis(
            rotation: float, oav_mode: Source
        ):
            yield from _rotate_in_one_direction_and_stream_to_redis(
                smargon, oav_to_redis_forwarder, oav_mode, rotation
            )

        yield from rotate_in_one_direction_and_stream_to_redis(
            rotation, Source.FULL_SCREEN
        )
        yield from rotate_in_one_direction_and_stream_to_redis(-rotation, Source.ROI)

    yield from do_thaw_and_stream_to_redis()


def _rotate_in_one_direction_and_stream_to_redis(
    smargon: Smargon,
    oav_to_redis_forwarder: OAVToRedisForwarder,
    oav_mode: Source,
    rotation: float,
):
    def get_metadata_from_current_oav():
        current_source_idx = yield from bps.rd(oav_to_redis_forwarder.selected_source)
        oav = oav_to_redis_forwarder.sources[current_source_idx].oav_ref()
        yield from bps.create()
        oav_info = yield from bps.read(oav)
        LOGGER.info(f"Got oav information: {oav_info}")
        yield from bps.save()

    yield from bps.mv(
        oav_to_redis_forwarder.selected_source,
        oav_mode.value,
    )

    yield from get_metadata_from_current_oav()
    yield from bps.monitor(smargon.omega.user_readback, name="gonio")
    yield from bps.monitor(oav_to_redis_forwarder.uuid, name="oav")

    yield from bps.kickoff(oav_to_redis_forwarder, wait=True)
    yield from bps.rel_set(smargon.omega, rotation, wait=True)
    yield from bps.complete(oav_to_redis_forwarder, wait=True)



from threading import Thread

from flask import Flask
from flask_restful import Api, Resource

from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.plan_runner import PlanRunner


# Ignore this function for code coverage as there is no way to shut down
# a server once it is started.
def create_server_for_udc(runner: PlanRunner, port: int) -> Thread:  # pragma: no cover
    """Create a minimal API for Hyperion UDC mode"""
    app = create_app_for_udc(runner)

    flask_thread = Thread(
        target=app.run,
        kwargs={"host": "0.0.0.0", "port": port},
        daemon=True,
    )
    flask_thread.start()
    LOGGER.info(f"Hyperion now listening on {port}")
    return flask_thread


def create_app_for_udc(runner: PlanRunner):
    app = Flask(__name__)
    api = Api(app)
    api.add_resource(StatusResource, "/status", resource_class_args=[runner])
    api.add_resource(CallbackLiveness, "/callbackPing", resource_class_args=[runner])
    return app


class StatusResource(Resource):
    """Status endpoint, used by k8s healthcheck probe"""

    def __init__(self, runner: PlanRunner):
        super().__init__()
        self._runner = runner

    def get(self):
        status = self._runner.current_status
        return {"status": status.value}


class CallbackLiveness(Resource):
    """Called periodically by the external callbacks to indicate that they are still running"""

    def __init__(self, runner: PlanRunner):
        super().__init__()
        self._runner = runner

    def get(self):
        LOGGER.debug("External callback ping received.")
        self._runner.reset_callback_watchdog_timer()



import threading
import time
from abc import abstractmethod
from collections.abc import Sequence

from blueapi.core import BlueskyContext
from bluesky import plan_stubs as bps
from bluesky.callbacks.zmq import Publisher
from bluesky.utils import MsgGenerator
from pydantic import BaseModel

from mx_bluesky.common.external_interaction.callbacks.common.log_uid_tag_callback import (
    LogUidTaggingCallback,
)
from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.parameters.constants import CONST


class PlanError(Exception):
    """Identifies an exception that was encountered during plan execution."""

    pass


class PlanRunner:
    EXTERNAL_CALLBACK_POLL_INTERVAL_S = 1
    EXTERNAL_CALLBACK_WATCHDOG_TIMER_S = 60

    def __init__(self, context: BlueskyContext, dev_mode: bool):
        self.context: BlueskyContext = context
        self.run_engine = context.run_engine
        # These references are necessary to maintain liveness of callbacks because run_engine
        # only keeps a weakref
        self._logging_uid_tag_callback = LogUidTaggingCallback()
        self._publisher = Publisher(f"localhost:{CONST.CALLBACK_0MQ_PROXY_PORTS[0]}")

        self.run_engine.subscribe(self._logging_uid_tag_callback)
        LOGGER.info("Connecting to external callback ZMQ proxy...")
        self.run_engine.subscribe(self._publisher)
        self._callbacks_started = False
        self._callback_watchdog_expiry = time.monotonic()
        self.is_dev_mode = dev_mode

    @abstractmethod
    def decode_and_execute(
        self, current_visit: str | None, parameter_list: Sequence[BaseModel]
    ) -> MsgGenerator:
        pass

    def reset_callback_watchdog_timer(self):
        """Called periodically to reset the watchdog timer when the external callbacks ping us."""
        self._callbacks_started = True
        self._callback_watchdog_expiry = (
            time.monotonic() + self.EXTERNAL_CALLBACK_WATCHDOG_TIMER_S
        )

    @property
    @abstractmethod
    def current_status(self) -> Status:
        pass

    @abstractmethod
    def shutdown(self):
        """Performs orderly prompt shutdown.
        Aborts the run engine and terminates the loop waiting for messages."""
        pass

    def check_external_callbacks_are_alive(self):
        callback_expiry = time.monotonic() + self.EXTERNAL_CALLBACK_WATCHDOG_TIMER_S
        while time.monotonic() < callback_expiry:
            if self._callbacks_started:
                break
            # If on first launch the external callbacks aren't started yet, wait until they are
            LOGGER.info("Waiting for external callbacks to start")
            yield from bps.sleep(self.EXTERNAL_CALLBACK_POLL_INTERVAL_S)
        else:
            raise RuntimeError("External callbacks not running - try restarting")

        if not self._external_callbacks_are_alive():
            raise RuntimeError(
                "External callback watchdog timer expired, check external callbacks are running."
            )

    def request_run_engine_abort(self):
        """Asynchronously request an abort from the run engine. This cannot be done from
        inside the main thread."""

        def issue_abort():
            try:
                # abort() causes the run engine to throw a RequestAbort exception
                # inside the plan, which will propagate through the contingency wrappers.
                # When the plan returns, the run engine will raise RunEngineInterrupted
                self.run_engine.abort()
            except Exception as e:
                LOGGER.warning(
                    "Exception encountered when issuing abort() to RunEngine:",
                    exc_info=e,
                )

        stopping_thread = threading.Thread(target=issue_abort)
        stopping_thread.start()

    def _external_callbacks_are_alive(self) -> bool:
        return time.monotonic() < self._callback_watchdog_expiry



from concurrent.futures import ThreadPoolExecutor
from threading import Event
from time import sleep
from unittest.mock import ANY, MagicMock, patch

import pytest
from blueapi.core import BlueskyContext
from bluesky import RunEngine
from bluesky import plan_stubs as bps

from mx_bluesky.common.utils.exceptions import CrystalNotFoundError
from mx_bluesky.hyperion.blueapi.parameters import LoadCentreCollectParams
from mx_bluesky.hyperion.in_process_runner import InProcessRunner
from mx_bluesky.hyperion.plan_runner import PlanError


@pytest.fixture(autouse=True)
def patch_timer_poll_interval():
    with patch(
        "mx_bluesky.hyperion.plan_runner.PlanRunner.EXTERNAL_CALLBACK_POLL_INTERVAL_S",
        0.01,
    ):
        yield


@pytest.fixture()
def patch_timer_expiry():
    with patch(
        "mx_bluesky.hyperion.plan_runner.PlanRunner.EXTERNAL_CALLBACK_WATCHDOG_TIMER_S",
        0.1,
    ):
        yield


def test_external_callbacks_waits_for_external_callback_ping(run_engine: RunEngine):
    runner = InProcessRunner(BlueskyContext(run_engine=run_engine), True)
    plan_started = Event()

    def execute_test():
        sleep(0.1)
        assert not plan_started.is_set()
        runner.reset_callback_watchdog_timer()
        plan_started.wait(timeout=0.5)

    def test_plan():
        plan_started.set()
        yield from bps.null()

    with ThreadPoolExecutor(1) as executor:
        fut = executor.submit(execute_test)
        run_engine(runner.execute_plan(test_plan))
        fut.result()


def test_external_callbacks_raises_if_never_started(
    run_engine: RunEngine, patch_timer_expiry
):
    runner = InProcessRunner(BlueskyContext(run_engine=run_engine), True)
    plan_started = Event()

    def execute_test():
        sleep(0.1)
        assert not plan_started.is_set()
        plan_started.wait(timeout=0.5)

    def test_plan():
        plan_started.set()
        yield from bps.null()

    with ThreadPoolExecutor(1) as executor:
        fut = executor.submit(execute_test)
        with pytest.raises(PlanError) as exc_info:
            run_engine(runner.execute_plan(test_plan))
        fut.result()

    assert exc_info.value.__cause__.args[0].startswith(  # type: ignore
        "External callbacks not running"
    )


def test_external_callbacks_not_running_raises_exception_for_plan_execution(
    run_engine: RunEngine,
    patch_timer_expiry,
):
    runner = InProcessRunner(BlueskyContext(run_engine=run_engine), True)

    def execute_test():
        runner.reset_callback_watchdog_timer()

    def test_plan():
        yield from bps.sleep(0.2)

    with ThreadPoolExecutor(1) as executor:
        fut = executor.submit(execute_test)
        run_engine(runner.execute_plan(test_plan))
        with pytest.raises(PlanError) as exc_info:
            run_engine(runner.execute_plan(test_plan))
        assert exc_info.value.__cause__.args[0].startswith(  # type:ignore
            "External callback watchdog timer expired"
        )
        fut.result()


@patch("mx_bluesky.hyperion.in_process_runner.create_devices", new=MagicMock())
@patch("mx_bluesky.hyperion.in_process_runner.load_centre_collect")
def test_in_process_runner_skips_native_collection_if_sample_error(
    mock_load_centre_collect: MagicMock,
    run_engine: RunEngine,
    external_load_centre_collect_params: LoadCentreCollectParams,
):
    mock_load_centre_collect.side_effect = CrystalNotFoundError(
        "Simulated crystal not found"
    )
    runner = InProcessRunner(BlueskyContext(run_engine=run_engine), True)
    runner.reset_callback_watchdog_timer()
    run_engine(
        runner.decode_and_execute(
            "TEST_VISIT",
            [external_load_centre_collect_params, external_load_centre_collect_params],
        )
    )
    mock_load_centre_collect.assert_called_once_with(ANY, ANY)



from __future__ import annotations

import dataclasses
from collections.abc import Sequence

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import numpy as np
from bluesky.utils import MsgGenerator, make_decorator
from dodal.common.beamlines.commissioning_mode import read_commissioning_mode
from dodal.devices.zocalo.zocalo_results import (
    XrcResult,
    ZocaloResults,
    get_full_processing_results,
)

from mx_bluesky.common.experiment_plans.inner_plans.do_fgs import ZOCALO_STAGE_GROUP
from mx_bluesky.common.parameters.constants import (
    GridscanParamConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundError,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.xrc_result import XRayCentreResult


def fetch_xrc_results_from_zocalo(
    zocalo_results: ZocaloResults,
    parameters: SpecifiedThreeDGridScan,
) -> MsgGenerator:
    """
    Get XRC results from the ZocaloResults device which was staged during a grid scan,
    and store them in XRayCentreEventHandler.xray_centre_results by firing an event.

    The RunEngine must be subscribed to XRayCentreEventHandler for this plan to work.
    """

    LOGGER.info("Getting X-ray center Zocalo results...")

    yield from bps.trigger(zocalo_results, wait=True)
    LOGGER.info("Zocalo triggered and read, interpreting results.")
    xrc_results = yield from get_full_processing_results(zocalo_results)
    LOGGER.info(f"Got xray centres, top 5: {xrc_results[:5]}")
    filtered_results = [
        result
        for result in xrc_results
        if result["total_count"]
        >= GridscanParamConstants.ZOCALO_MIN_TOTAL_COUNT_THRESHOLD
    ]
    discarded_count = len(xrc_results) - len(filtered_results)
    if discarded_count > 0:
        LOGGER.info(f"Removed {discarded_count} results because below threshold")
    if filtered_results:
        flyscan_results = [
            _xrc_result_in_boxes_to_result_in_mm(xr, parameters)
            for xr in filtered_results
        ]
    else:
        commissioning_mode = yield from read_commissioning_mode()
        if commissioning_mode:
            LOGGER.info("Commissioning mode enabled, returning dummy result")
            flyscan_results = [_generate_dummy_xrc_result(parameters)]
        else:
            LOGGER.warning("No X-ray centre received")
            raise CrystalNotFoundError()
    yield from _fire_xray_centre_result_event(flyscan_results)


def _generate_dummy_xrc_result(params: SpecifiedThreeDGridScan) -> XRayCentreResult:
    com = [params.x_steps / 2, params.y_steps / 2, params.z_steps / 2]
    max_voxel = [round(p) for p in com]
    return _xrc_result_in_boxes_to_result_in_mm(
        XrcResult(
            centre_of_mass=com,
            max_voxel=max_voxel,
            bounding_box=[max_voxel, [p + 1 for p in max_voxel]],
            n_voxels=1,
            max_count=10000,
            total_count=100000,
            sample_id=params.sample_id,
        ),
        params,
    )


def _xrc_result_in_boxes_to_result_in_mm(
    xrc_result: XrcResult, parameters: SpecifiedThreeDGridScan
) -> XRayCentreResult:
    fgs_params = parameters.fast_gridscan_params
    xray_centre = fgs_params.grid_position_to_motor_position(
        np.array(xrc_result["centre_of_mass"])
    )
    # A correction is applied to the bounding box to map discrete grid coordinates to
    # the corners of the box in motor-space; we do not apply this correction
    # to the xray-centre as it is already in continuous space and the conversion has
    # been performed already
    # In other words, xrc_result["bounding_box"] contains the position of the box centre,
    # so we subtract half a box to get the corner of the box
    return XRayCentreResult(
        centre_of_mass_mm=xray_centre,
        bounding_box_mm=(
            fgs_params.grid_position_to_motor_position(
                np.array(xrc_result["bounding_box"][0]) - 0.5
            ),
            fgs_params.grid_position_to_motor_position(
                np.array(xrc_result["bounding_box"][1]) - 0.5
            ),
        ),
        max_count=xrc_result["max_count"],
        total_count=xrc_result["total_count"],
        sample_id=xrc_result["sample_id"],
    )


def _fire_xray_centre_result_event(results: Sequence[XRayCentreResult]):
    def empty_plan():
        return iter([])

    yield from bpp.set_run_key_wrapper(
        bpp.run_wrapper(
            empty_plan(),
            md={
                PlanNameConstants.FLYSCAN_RESULTS: [
                    dataclasses.asdict(r) for r in results
                ]
            },
        ),
        PlanNameConstants.FLYSCAN_RESULTS,
    )


# Remove after https://github.com/bluesky/bluesky/issues/1979
def _zocalo_stage_wrapper(plan: MsgGenerator, zocalo: ZocaloResults):
    yield from bps.stage(zocalo, group=ZOCALO_STAGE_GROUP)
    yield from bpp.contingency_wrapper(
        plan, final_plan=lambda: (yield from bps.unstage(zocalo))
    )


zocalo_stage_decorator = make_decorator(_zocalo_stage_wrapper)



from enum import StrEnum

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aithre_lasershaping.goniometer import Goniometer


class JogDirection(StrEnum):
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"
    ZPLUS = "z_plus"
    ZMINUS = "z_minus"


def rotate_goniometer_relative(
    value: float, goniometer: Goniometer = inject("goniometer")
) -> MsgGenerator:
    """Adjust the goniometer position incrementally"""
    yield from bps.rel_set(goniometer.omega, value, wait=True)


def change_goniometer_turn_speed(
    velocity: float, goniometer: Goniometer = inject("goniometer")
) -> MsgGenerator:
    """Set the velocity of the goniometer"""
    yield from bps.mv(goniometer.omega.velocity, velocity)


def jog_sample(
    direction: JogDirection,
    increment_size: float,
    goniometer: Goniometer = inject("goniometer"),
) -> MsgGenerator:
    """Adjust the goniometer stage positions"""
    direction_map = {
        JogDirection.RIGHT: (goniometer.x, 1),
        JogDirection.LEFT: (goniometer.x, -1),
        JogDirection.ZPLUS: (goniometer.z, 1),
        JogDirection.ZMINUS: (goniometer.z, -1),
        JogDirection.UP: (goniometer.vertical_position, 1),
        JogDirection.DOWN: (goniometer.vertical_position, -1),
    }

    axis, sign = direction_map[direction]
    yield from bps.mvr(axis, sign * increment_size)


def go_to_furthest_maximum(
    goniometer: Goniometer = inject("goniometer"),
) -> MsgGenerator:
    """Rotate to positive or negative maximum, whichever is further away"""

    limit_of_travel = 3600
    current_value: float = yield from bps.rd(goniometer.omega.user_readback)

    yield from bps.mv(
        goniometer.omega, -limit_of_travel if current_value > 0 else limit_of_travel
    )



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from pydantic import BaseModel

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    get_proposal_and_session_from_visit_string,
)
from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGroupInfo,
    DataCollectionInfo,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import ExpeyeInteraction
from mx_bluesky.common.external_interaction.ispyb.ispyb_utils import (
    get_current_time_string,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMadeError
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER

if TYPE_CHECKING:
    pass


class IspybIds(BaseModel):
    data_collection_ids: tuple[int, ...] = ()
    data_collection_group_id: int | None = None
    grid_ids: tuple[int, ...] = ()


class StoreInIspyb:
    def __init__(self, ispyb_config: str) -> None:
        self.ISPYB_CONFIG_PATH: str = ispyb_config
        self._expeye = ExpeyeInteraction()

    def begin_deposition(
        self,
        data_collection_group_info: DataCollectionGroupInfo,
        scan_data_infos: Sequence[ScanDataInfo],
    ) -> IspybIds:
        ispyb_ids = IspybIds()
        if scan_data_infos[0].data_collection_info:
            ispyb_ids.data_collection_group_id = scan_data_infos[
                0
            ].data_collection_info.parent_id

        return self._begin_or_update_deposition(
            ispyb_ids, data_collection_group_info, scan_data_infos
        )

    def update_deposition(
        self,
        ispyb_ids,
        scan_data_infos: Sequence[ScanDataInfo],
    ) -> IspybIds:
        assert ispyb_ids.data_collection_group_id, (
            "Attempted to store scan data without a collection group"
        )
        assert ispyb_ids.data_collection_ids, (
            "Attempted to store scan data without a collection"
        )
        return self._begin_or_update_deposition(ispyb_ids, None, scan_data_infos)

    def _begin_or_update_deposition(
        self,
        ispyb_ids,
        data_collection_group_info: DataCollectionGroupInfo | None,
        scan_data_infos,
    ) -> IspybIds:
        if data_collection_group_info:
            ispyb_ids.data_collection_group_id = (
                self._store_data_collection_group_table(
                    data_collection_group_info, ispyb_ids.data_collection_group_id
                )
            )
        else:
            assert ispyb_ids.data_collection_group_id, (
                "Attempt to update data collection without a data collection group ID"
            )

        grid_ids = list(ispyb_ids.grid_ids)
        data_collection_ids_out = list(ispyb_ids.data_collection_ids)
        for scan_data_info in scan_data_infos:
            data_collection_id = scan_data_info.data_collection_id
            if (
                scan_data_info.data_collection_info
                and not scan_data_info.data_collection_info.parent_id
            ):
                scan_data_info.data_collection_info.parent_id = (
                    ispyb_ids.data_collection_group_id
                )

            new_data_collection_id, grid_id = self._store_single_scan_data(
                scan_data_info, data_collection_id
            )
            if not data_collection_id:
                data_collection_ids_out.append(new_data_collection_id)
            if grid_id:
                grid_ids.append(grid_id)
        ispyb_ids = IspybIds(
            data_collection_ids=tuple(data_collection_ids_out),
            grid_ids=tuple(grid_ids),
            data_collection_group_id=ispyb_ids.data_collection_group_id,
        )
        return ispyb_ids

    def end_deposition(self, ispyb_ids: IspybIds, success: str, reason: str):
        assert ispyb_ids.data_collection_ids, (
            "Can't end ISPyB deposition, data_collection IDs are missing"
        )
        assert ispyb_ids.data_collection_group_id is not None, (
            "Cannot end ISPyB deposition without data collection group ID"
        )

        for id_ in ispyb_ids.data_collection_ids:
            ISPYB_ZOCALO_CALLBACK_LOGGER.info(
                f"End ispyb deposition with status '{success}' and reason '{reason}'."
            )
            if success == "fail" or success == "abort":
                run_status = "DataCollection Unsuccessful"
            else:
                run_status = "DataCollection Successful"
            current_time = get_current_time_string()
            self._update_scan_with_end_time_and_status(
                current_time, run_status, reason, id_
            )

    def append_to_comment(
        self, data_collection_id: int, comment: str, delimiter: str = " "
    ) -> None:
        try:
            self._expeye.update_data_collection(
                data_collection_id,
                DataCollectionInfo(comments=delimiter + comment),
                True,
            )
        except ISPyBDepositionNotMadeError as e:
            ISPYB_ZOCALO_CALLBACK_LOGGER.warning(
                f"Unable to log comment, comment probably exceeded column length: {comment}",
                exc_info=e,
            )

    def update_data_collection_group_table(
        self,
        dcg_info: DataCollectionGroupInfo,
        data_collection_group_id: int | None = None,
    ) -> None:
        self._store_data_collection_group_table(dcg_info, data_collection_group_id)

    def _update_scan_with_end_time_and_status(
        self, end_time: str, run_status: str, reason: str, data_collection_id: int
    ) -> None:
        if reason != "":
            self.append_to_comment(data_collection_id, f"{run_status} reason: {reason}")

        info = DataCollectionInfo(end_time=end_time, run_status=run_status)
        self._expeye.update_data_collection(data_collection_id, info)

    def _store_data_collection_group_table(
        self,
        dcg_info: DataCollectionGroupInfo,
        data_collection_group_id: int | None = None,
    ) -> int:
        if data_collection_group_id:
            self._expeye.update_data_group(data_collection_group_id, dcg_info)
            return data_collection_group_id
        else:
            proposal, session = get_proposal_and_session_from_visit_string(
                dcg_info.visit_string
            )
            return self._expeye.create_data_group(proposal, session, dcg_info)

    def _store_data_collection_table(
        self, data_collection_id, data_collection_info: DataCollectionInfo
    ) -> int:
        if data_collection_id and data_collection_info.comments:
            self.append_to_comment(
                data_collection_id, data_collection_info.comments, " "
            )
            data_collection_info.comments = None

        if data_collection_id:
            self._expeye.update_data_collection(
                data_collection_id, data_collection_info
            )
            return data_collection_id
        else:
            assert data_collection_info.parent_id, (
                "Data Collection must have a Data Collection Group"
            )
            return self._expeye.create_data_collection(
                data_collection_info.parent_id, data_collection_info
            )

    def _store_single_scan_data(
        self, scan_data_info, data_collection_id=None
    ) -> tuple[int, int | None]:
        data_collection_id = self._store_data_collection_table(
            data_collection_id, scan_data_info.data_collection_info
        )

        if scan_data_info.data_collection_position_info:
            self._expeye.create_position(
                data_collection_id, scan_data_info.data_collection_position_info
            )

        grid_id = None
        if scan_data_info.data_collection_grid_info:
            grid_id = self._expeye.create_grid(
                data_collection_id, scan_data_info.data_collection_grid_info
            )
        return data_collection_id, grid_id



from __future__ import annotations

import os
import signal
import threading
from pathlib import Path
from sys import argv
from time import sleep
from unittest.mock import ANY, MagicMock, call, patch

import pytest
from blueapi.config import ApplicationConfig
from dodal.devices.baton import Baton
from dodal.devices.synchrotron import Synchrotron
from ophyd_async.core import set_mock_value

from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.utils.context import (
    find_device_in_context,
)
from mx_bluesky.hyperion.__main__ import (
    initialise_globals,
    main,
)
from mx_bluesky.hyperion.baton_handler import HYPERION_USER
from mx_bluesky.hyperion.parameters.cli import (
    HyperionArgs,
    HyperionMode,
    parse_cli_args,
)
from mx_bluesky.hyperion.parameters.constants import CONST, HyperionConstants
from mx_bluesky.hyperion.plan_runner import PlanRunner

from .conftest import AGAMEMNON_WAIT_INSTRUCTION


@pytest.fixture(autouse=True)
def patch_remote_graylog_endpoint():
    with patch("dodal.log.get_graylog_configuration", return_value=("localhost", 5555)):
        yield None


@pytest.fixture(autouse=True)
def mock_create_udc_server():
    with patch("mx_bluesky.hyperion.__main__.create_server_for_udc") as mock_udc_server:
        yield mock_udc_server


@pytest.fixture
def mock_setup_context(request: pytest.FixtureRequest):
    with (
        patch("mx_bluesky.hyperion.__main__.setup_context") as mock_setup_context,
        patch("mx_bluesky.hyperion.__main__.run_forever"),
    ):
        yield mock_setup_context


@pytest.mark.parametrize(
    ["arg_list", "parsed_arg_values"],
    [
        (
            [
                "--dev",
            ],
            (True,),
        ),
        ([], (False,)),
    ],
)
def test_cli_args_parse(arg_list, parsed_arg_values):
    argv[1:] = arg_list
    test_args = parse_cli_args()
    assert test_args.dev_mode == parsed_arg_values[0]


@pytest.fixture(autouse=True)
def beamline_i03():
    with (
        patch.dict(os.environ, {"BEAMLINE": "i03"}),
        patch.dict(
            "dodal.common.beamlines.beamline_parameters.BEAMLINE_PARAMETER_PATHS",
            {"i03": "tests/test_data/test_beamline_parameters.txt"},
        ),
    ):
        yield


@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
@pytest.mark.parametrize("dev_mode", [False, True])
def test_context_created_with_dev_mode(dev_mode: bool, mock_setup_context: MagicMock):
    with (
        patch("sys.argv", new=["hyperion", "--dev"] if dev_mode else ["hyperion"]),
        patch("mx_bluesky.hyperion.__main__.create_server_for_udc"),
    ):
        main()

    mock_setup_context.assert_called_once_with(dev_mode=dev_mode)


@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup")
@patch("mx_bluesky.hyperion.__main__.alerting.set_alerting_service")
def test_initialise_configures_logging(
    mock_alerting_setup: MagicMock, mock_logging_setup: MagicMock
):
    args = HyperionArgs(mode=HyperionMode.UDC, dev_mode=True)

    initialise_globals(args)

    mock_logging_setup.assert_called_once_with(
        CONST.LOG_FILE_NAME, CONST.GRAYLOG_PORT, dev_mode=True
    )


@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup")
@patch("mx_bluesky.hyperion.__main__.alerting.set_alerting_service")
def test_initialise_configures_alerting(
    mock_alerting_setup: MagicMock, mock_logging_setup: MagicMock
):
    args = HyperionArgs(mode=HyperionMode.UDC, dev_mode=True)

    initialise_globals(args)

    mock_alerting_setup.assert_called_once()
    assert isinstance(mock_alerting_setup.mock_calls[0].args[0], LoggingAlertService)


@patch("sys.argv", new=["hyperion", "--mode", "udc"])
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup")
@patch("mx_bluesky.hyperion.__main__.create_server_for_udc", MagicMock())
@patch("mx_bluesky.hyperion.__main__.run_forever", MagicMock())
def test_hyperion_in_udc_mode_starts_logging(
    mock_do_default_logging_setup: MagicMock,
    mock_setup_context: MagicMock,
):
    main()

    mock_do_default_logging_setup.assert_called_once_with(
        CONST.LOG_FILE_NAME, CONST.GRAYLOG_PORT, dev_mode=False
    )


@patch("sys.argv", new=["hyperion", "--mode", "udc"])
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
@patch("mx_bluesky.hyperion.__main__.run_forever", MagicMock())
def test_hyperion_in_udc_mode_starts_udc_api_on_hyperion_port(
    mock_create_udc_server: MagicMock,
    mock_setup_context: MagicMock,
):
    main()
    mock_create_udc_server.assert_called_once_with(ANY, HyperionConstants.HYPERION_PORT)
    assert isinstance(mock_create_udc_server.mock_calls[0].args[0], PlanRunner)


@patch("sys.argv", new=["hyperion", "--mode", "udc"])
@patch("mx_bluesky.hyperion.__main__.setup_context")
@patch("mx_bluesky.hyperion.__main__.run_forever")
@patch("mx_bluesky.hyperion.baton_handler.find_device_in_context", autospec=True)
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
def test_hyperion_in_udc_mode_starts_udc_loop(
    mock_find_device_in_context: MagicMock,
    mock_run_forever: MagicMock,
    mock_setup_context: MagicMock,
):
    main()

    mock_run_forever.assert_called_once()
    assert isinstance(mock_run_forever.mock_calls[0].args[0], PlanRunner)


@patch(
    "sys.argv",
    new=["hyperion", "--mode", "supervisor", "--supervisor-config", "test_config"],
)
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
def test_hyperion_in_supervisor_mode_requires_client_config_option():
    with pytest.raises(
        RuntimeError,
        match="BlueAPI client configuration file must be specified in supervisor mode.",
    ):
        main()


@patch(
    "sys.argv",
    new=["hyperion", "--mode", "supervisor", "--client-config", "test_config"],
)
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
def test_hyperion_in_supervisor_mode_requires_supervisor_config_option():
    with pytest.raises(
        RuntimeError,
        match="BlueAPI supervisor configuration file must be specified in supervisor mode.",
    ):
        main()


@pytest.fixture
def mock_supervisor_mode():
    parent = MagicMock()
    with patch.multiple(
        "mx_bluesky.hyperion.__main__",
        ConfigLoader=parent.ConfigLoader,
        BlueskyContext=parent.BlueskyContext,
        run_forever=parent.run_forever,
        signal=parent.signal,
        SupervisorRunner=parent.SupervisorRunner,
    ):
        yield parent


@patch(
    "sys.argv",
    new=[
        "hyperion",
        "--mode",
        "supervisor",
        "--client-config",
        "test_client_config",
        "--supervisor-config",
        "test_supervisor_config",
    ],
)
@patch("mx_bluesky.hyperion.__main__.run_forever", MagicMock())
@patch("mx_bluesky.hyperion.__main__.do_default_logging_setup", MagicMock())
def test_hyperion_in_supervisor_mode_creates_rest_server_on_supervisor_port(
    mock_supervisor_mode: MagicMock,
    mock_create_udc_server: MagicMock,
):
    mock_supervisor_mode.ConfigLoader.return_value.load.side_effect = [
        "client_config",
        "supervisor_config",
    ]
    main()
    mock_supervisor_mode.assert_has_calls(
        [
            call.ConfigLoader(ApplicationConfig),
            call.ConfigLoader().use_values_from_yaml(Path("test_client_config")),
            call.ConfigLoader().load(),
            call.ConfigLoader(ApplicationConfig),
            call.ConfigLoader().use_values_from_yaml(Path("test_supervisor_config")),
            call.ConfigLoader().load(),
            call.BlueskyContext(configuration="supervisor_config"),
            call.SupervisorRunner(ANY, "client_config", False),
        ]
    )
    mock_create_udc_server.assert_called_once_with(
        ANY, HyperionConstants.SUPERVISOR_PORT
    )


@patch("sys.argv", new=["hyperion", "--mode", "udc", "--dev"])
@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    return_value=[AGAMEMNON_WAIT_INSTRUCTION],
)
@patch("mx_bluesky.hyperion.baton_handler.clear_all_device_caches", MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.setup_devices", MagicMock())
def test_sending_main_process_sigterm_in_udc_mode_performs_clean_prompt_shutdown(
    mock_create_parameters_from_agamemnon,
    use_beamline_t01,
    mock_create_udc_server,
):
    def wait_for_udc_to_start_then_send_sigterm():
        while len(mock_create_udc_server.mock_calls) == 0:
            sleep(0.2)

        plan_runner = mock_create_udc_server.mock_calls[0].args[0]
        context = plan_runner.context
        baton = find_device_in_context(context, "baton", Baton)
        synchrotron = find_device_in_context(context, "synchrotron", Synchrotron)
        set_mock_value(baton.requested_user, HYPERION_USER)
        set_mock_value(synchrotron.machine_user_countdown, 1200)
        while len(mock_create_parameters_from_agamemnon.mock_calls) == 0:
            sleep(0.2)
        os.kill(os.getpid(), signal.SIGTERM)

    t = threading.Thread(None, wait_for_udc_to_start_then_send_sigterm, daemon=True)
    t.start()
    main()



"""
This package contains the bluesky plan entry points for use with hyperion-blueapi.
The json schema and documentation therein generated by the blueapi /plans endpoint
from this package constitutes the hyperion-blueapi interface to the hyperion supervisor
process.
"""



import json
import logging
from logging.handlers import TimedRotatingFileHandler
from os import environ
from pathlib import Path

from dodal.log import (
    ERROR_LOG_BUFFER_LINES,
    CircularMemoryHandler,
    DodalLogHandlers,
    integrate_bluesky_and_ophyd_logging,
    set_up_all_logging_handlers,
    set_up_debug_memory_handler,
    set_up_info_file_handler,
)
from dodal.log import LOGGER as DODAL_LOGGER

LOGGER = logging.getLogger("MX-Bluesky")
LOGGER.setLevel("DEBUG")
LOGGER.parent = DODAL_LOGGER

ISPYB_ZOCALO_CALLBACK_LOGGER = logging.getLogger("ISPyB and Zocalo callbacks")
ISPYB_ZOCALO_CALLBACK_LOGGER.setLevel(logging.DEBUG)

NEXUS_LOGGER = logging.getLogger("NeXus callbacks")
NEXUS_LOGGER.setLevel(logging.DEBUG)

ALL_LOGGERS = [LOGGER, ISPYB_ZOCALO_CALLBACK_LOGGER, NEXUS_LOGGER]

__logger_handlers: DodalLogHandlers | None = None


def format_doc_for_log(doc):
    class _BestEffortEncoder(json.JSONEncoder):
        def default(self, o):
            return repr(o)

    return json.dumps(doc, indent=2, cls=_BestEffortEncoder)


class ExperimentMetadataTagFilter(logging.Filter):
    """When an instance of this custom filter is added to a logging handler, dc_group_id
    and run_id will be tagged in that handlers' log messages."""

    dc_group_id: str | None = None
    run_uid: str | None = None

    def filter(self, record):
        if self.dc_group_id:
            record.dc_group_id = self.dc_group_id
        if self.run_uid:
            record.run_uid = self.run_uid
        return True


tag_filter = ExperimentMetadataTagFilter()


def set_dcgid_tag(dcgid):
    """Set the datacollection group id as a tag on all subsequent log messages.
    Setting to None will remove the tag."""
    tag_filter.dc_group_id = dcgid


def set_uid_tag(uid):
    """Set the unique id as a tag on all subsequent log messages.
    Setting to None will remove the tag."""
    tag_filter.run_uid = uid


def setup_hyperion_blueapi_logging(log_file_name: str):
    """Configure debug logging for hyperion-blueapi.
    Args:
        log_file_name: Base name of the log file.
    """
    dodal_logger = DODAL_LOGGER
    logging_path, debug_logging_path = _get_logging_dirs(False)
    set_up_debug_memory_handler(
        dodal_logger, debug_logging_path, log_file_name, ERROR_LOG_BUFFER_LINES
    )
    set_up_info_file_handler(
        dodal_logger,
        logging_path,
        log_file_name,
    )


def do_default_logging_setup(
    file_name: str,
    graylog_port: int,
    dev_mode: bool = False,
    integrate_all_logs: bool = True,
):
    """Configures dodal logger so that separate debug and info log files are created,
    info logs are sent to Graylog, info logs are streamed to sys.sterr, and logs from ophyd
    and bluesky and ophyd-async are optionally included."""
    logging_path, debug_logging_path = _get_logging_dirs(dev_mode)
    handlers = set_up_all_logging_handlers(
        DODAL_LOGGER,
        logging_path,
        file_name,
        dev_mode,
        ERROR_LOG_BUFFER_LINES,
        graylog_port,
        debug_logging_path,
    )

    if integrate_all_logs:
        integrate_bluesky_and_ophyd_logging(DODAL_LOGGER)

    handlers["graylog_handler"].addFilter(tag_filter)

    global __logger_handlers
    __logger_handlers = handlers


def _get_debug_handler() -> CircularMemoryHandler:
    assert __logger_handlers is not None, (
        "You can only use this after running the default logging setup"
    )
    return __logger_handlers["debug_memory_handler"]


def flush_debug_handler() -> str:
    """Writes the contents of the circular debug log buffer to disk and returns the written filename"""
    handler = _get_debug_handler()
    assert isinstance(handler.target, TimedRotatingFileHandler), (
        "Circular memory handler doesn't have an appropriate fileHandler target"
    )
    handler.flush()
    return handler.target.baseFilename


def _get_logging_dirs(dev_mode: bool) -> tuple[Path, Path]:
    """Get the paths to write the mx_bluesky log files to.

    Log location must be specified in the LOG_DIR environment variable,
    and the debug log location specified in the DEBUG_LOG_DIR environment variable.
    This directory will be created if it is not found.

    Logs are written to ./tmp/logs/bluesky if BEAMLINE environment variable is not found

    Args:
        dev_mode (bool): If True, the logs will be written to /tmp/logs/bluesky if the environment variables are not set
    Returns:
        tuple[Path, Path]: Paths to the standard log file and to the debug log file, for the file handlers to write to
    Raises:
        ValueError: If LOG_DIR or DEBUG_LOG_DIR environment variable is not set and dev_mode is False
    """
    if dev_mode:
        logging_path = Path(environ.get("LOG_DIR", "/tmp/logs/bluesky"))
        debug_logging_path = Path(environ.get("DEBUG_LOG_DIR", "/tmp/logs/bluesky"))
    else:
        logging_dir = environ.get("LOG_DIR")
        if not logging_dir:
            raise ValueError("LOG_DIR environment variable is not set")
        debug_logging_dir = environ.get("DEBUG_LOG_DIR")
        if not debug_logging_dir:
            raise ValueError("DEBUG_LOG_DIR environment variable is not set")
        logging_path = Path(logging_dir)
        debug_logging_path = Path(debug_logging_dir)

    Path.mkdir(logging_path, exist_ok=True, parents=True)
    Path.mkdir(debug_logging_path, exist_ok=True, parents=True)
    return logging_path, debug_logging_path



from enum import Enum

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.beamlines.i24.pmac import PMAC


class MoveSize(Enum):
    SMALL = "small"
    BIG = "big"


class Direction(Enum):
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"


class FocusDirection(Enum):
    IN = "in"
    OUT = "out"


def _move_direction(magnitude: float, direction: Direction, pmac):
    y_move = 0.0
    x_move = 0.0

    match direction:
        case Direction.UP:
            y_move = magnitude
        case Direction.DOWN:
            y_move = -magnitude
        case Direction.LEFT:
            x_move = -magnitude
        case Direction.RIGHT:
            x_move = magnitude

    yield from bps.abs_set(pmac.x, x_move, wait=True)
    yield from bps.abs_set(pmac.y, y_move, wait=True)


def move_block_on_arrow_click(direction: Direction, pmac: PMAC = inject("pmac")):
    magnitude = 3.1750
    yield from _move_direction(magnitude, direction, pmac)


def move_window_on_arrow_click(
    direction: Direction, size_of_move: MoveSize, pmac: PMAC = inject("pmac")
):
    match size_of_move:
        case MoveSize.SMALL:
            magnitude = 0.1250
        case MoveSize.BIG:
            magnitude = 0.3750

    yield from _move_direction(magnitude, direction, pmac)


def move_nudge_on_arrow_click(
    direction: Direction, size_of_move: MoveSize, pmac: PMAC = inject("pmac")
):
    match size_of_move:
        case MoveSize.SMALL:
            magnitude = 0.0010
        case MoveSize.BIG:
            magnitude = 0.0060

    yield from _move_direction(magnitude, direction, pmac)


def focus_on_oav_view(
    direction: FocusDirection, size_of_move: MoveSize, pmac: PMAC = inject("pmac")
) -> MsgGenerator:
    match size_of_move:
        case MoveSize.SMALL:
            magnitude = 0.0200
        case MoveSize.BIG:
            magnitude = 0.1200

    if direction == FocusDirection.IN:
        magnitude = -magnitude

    yield from bps.abs_set(pmac.z, magnitude, wait=True)



import io
import json
import logging
import pickle
from datetime import timedelta
from logging import StreamHandler
from typing import TypedDict

import numpy as np
import zmq
from dodal.devices.beamlines.i04.constants import RedisConstants
from dodal.devices.beamlines.i04.murko_results import (
    RESULTS_COMPLETE_MESSAGE,
    MurkoResult,
)
from numpy.typing import NDArray
from PIL import Image
from redis import StrictRedis

from mx_bluesky.beamlines.i04.callbacks.murko_callback import (
    FORWARDING_COMPLETE_MESSAGE,
)
from mx_bluesky.common.utils.log import LOGGER

MURKO_ADDRESS = "tcp://i04-murko-prod.diamond.ac.uk:8008"


FullMurkoResults = dict[str, list[MurkoResult]]


class MurkoRequest(TypedDict):
    """See https://github.com/MartinSavko/murko#usage for more information."""

    to_predict: NDArray
    model_img_size: tuple[int, int]
    save: bool
    min_size: int
    description: list

    # The identifier for each image
    prefix: list[str]


def get_image_size(image: NDArray) -> tuple[int, int]:
    """Returns the width and height of a numpy image"""
    return image.shape[1], image.shape[0]


def send_to_murko_and_get_results(request: MurkoRequest) -> FullMurkoResults:
    LOGGER.info(f"Sending {request['prefix']} to murko")
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect(MURKO_ADDRESS)
    socket.send(pickle.dumps(request))
    raw_results = socket.recv()
    assert isinstance(raw_results, bytes)
    results = pickle.loads(raw_results)
    LOGGER.info(f"Got {len(results['descriptions'])} results")
    return results


def _correlate_results_to_uuids(
    request: MurkoRequest, murko_results: FullMurkoResults
) -> list[tuple[str, MurkoResult]]:
    """We send a batch of images to murko, with each having a 'prefix' of the uuid that
    we're using to keep track of the image. Murko sends back an ordered list of these,
    which we match to the supplied prefix here."""
    return list(zip(request["prefix"], murko_results["descriptions"], strict=False))


class BatchMurkoForwarder:
    def __init__(self, redis_client: StrictRedis, batch_size: int):
        """
        Holds image data streamed from redis and forwards it to murko when:
            * A set number have been received
            * The shape of the images changes
            * When `flush` is called

        Once data has been forwarded this will then wait on the results and put them
        back in redis.

        Args:
            redis_client: The client to send murko results back to redis.
            batch_size: How many results to accumulate until they are flushed to redis.
        """
        self.redis_client = redis_client
        self.batch_size = batch_size
        self._uuids_and_images: dict[str, NDArray] = {}
        self._last_image_size: tuple[int, int] | None = None
        self._last_sample_id = ""

    def _handle_batch_of_images(self, sample_id, images, uuids):
        request_arguments: MurkoRequest = {
            "model_img_size": (256, 320),
            "to_predict": np.array(images),
            "save": False,
            "min_size": 64,
            "description": [
                "foreground",
                "crystal",
                "loop_inside",
                "loop",
                ["crystal", "loop"],
                ["crystal", "loop", "stem"],
            ],
            "prefix": uuids,
        }

        results = send_to_murko_and_get_results(request_arguments)
        results_with_uuids = _correlate_results_to_uuids(request_arguments, results)
        self._send_murko_results_to_redis(sample_id, results_with_uuids)

    def _send_murko_results_to_redis(
        self, sample_id: str, results: list[tuple[str, MurkoResult]]
    ):
        """Stores the results into a redis hash (for longer term storage) and publishes
        them as well so that downstream clients can get notified."""
        for uuid, result in results:
            redis_key = f"murko:{sample_id}:results"
            self.redis_client.hset(redis_key, uuid, str(pickle.dumps(result)))
            self.redis_client.expire(redis_key, timedelta(days=7))
        self.redis_client.publish("murko-results", pickle.dumps(results))

    def send_stop_message_to_redis(self):
        LOGGER.info(f"Publishing results complete message: {RESULTS_COMPLETE_MESSAGE}")
        self.redis_client.publish(
            "murko-results", pickle.dumps(RESULTS_COMPLETE_MESSAGE)
        )

    def add(self, sample_id: str, uuid: str, image: NDArray):
        """Add an image to the batch to send to murko."""
        image_size = get_image_size(image)
        self._last_sample_id = sample_id
        if self._last_image_size and self._last_image_size != image_size:
            self.flush()
        self._uuids_and_images[uuid] = image
        self._last_image_size = image_size
        if len(self._uuids_and_images.keys()) >= self.batch_size:
            self.flush()

    def flush(self):
        """Flush the batch to murko."""
        if self._uuids_and_images:
            self._handle_batch_of_images(
                self._last_sample_id,
                list(self._uuids_and_images.values()),
                list(self._uuids_and_images.keys()),
            )
        self._uuids_and_images = {}
        self._last_image_size = None


class RedisListener:
    TIMEOUT_S = 2

    def __init__(
        self,
        redis_host=RedisConstants.REDIS_HOST,
        redis_password=RedisConstants.REDIS_PASSWORD,
        db=RedisConstants.MURKO_REDIS_DB,
        redis_channel="murko",
    ):
        self.redis_client = StrictRedis(
            host=redis_host,
            password=redis_password,
            db=db,
        )
        self.pubsub = self.redis_client.pubsub()
        self.channel = redis_channel
        self.forwarder = BatchMurkoForwarder(self.redis_client, 10)

    def _get_and_handle_message(self):
        message = self.pubsub.get_message(timeout=self.TIMEOUT_S)
        if message and message["type"] == "message":
            assert message["data"] is not None
            data = json.loads(message["data"])
            LOGGER.info(f"Received from redis: {data}")
            if data == FORWARDING_COMPLETE_MESSAGE:
                LOGGER.info(
                    f"Received forwarding complete message: {FORWARDING_COMPLETE_MESSAGE}"
                )
                self.forwarder.flush()
                self.forwarder.send_stop_message_to_redis()
                return
            uuid = data["uuid"]
            sample_id = data["sample_id"]

            # Images are put in redis as raw jpeg bytes, murko needs numpy arrays
            image_key = f"murko:{sample_id}:raw"
            raw_image = self.redis_client.hget(image_key, uuid)

            if not isinstance(raw_image, bytes):
                LOGGER.warning(
                    f"Image at {image_key}:{uuid} is {raw_image}, expected bytes. Ignoring the data"
                )
                return

            image = Image.open(io.BytesIO(raw_image))
            image = np.asarray(image)

            self.forwarder.add(sample_id, uuid, image)

        elif not message:
            self.forwarder.flush()

    def listen_for_image_data_forever(self):
        self.pubsub.subscribe(self.channel)

        while True:
            self._get_and_handle_message()


def main():
    stream_handler = StreamHandler()
    stream_handler.setLevel(logging.INFO)
    LOGGER.addHandler(stream_handler)

    client = RedisListener()
    client.listen_for_image_data_forever()


if __name__ == "__main__":
    main()



import bluesky.plan_stubs as bps
import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import assert_message_and_return_remaining
from dodal.devices.zocalo import ZocaloResults

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    FlyScanEssentialDevices,
)
from mx_bluesky.common.experiment_plans.inner_plans.do_fgs import ZOCALO_STAGE_GROUP
from mx_bluesky.common.experiment_plans.inner_plans.xrc_results_utils import (
    fetch_xrc_results_from_zocalo,
    zocalo_stage_decorator,
)
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundError,
)
from mx_bluesky.common.utils.xrc_result import XRayCentreEventHandler, XRayCentreResult
from tests.conftest import (
    RunEngineSimulator,
)
from tests.unit_tests.hyperion.experiment_plans.conftest import (
    mock_zocalo_trigger,
)

from ....conftest import TestData

""" Below are tests for getting zocalo results from the zocalo device, which is done
 post gridscan """


async def test_results_adjusted_and_event_raised(
    test_fgs_params: SpecifiedThreeDGridScan,
    run_engine: RunEngine,
    zocalo: ZocaloResults,
):
    x_ray_centre_event_handler = XRayCentreEventHandler()
    run_engine.subscribe(x_ray_centre_event_handler)
    mock_zocalo_trigger(zocalo, TestData.test_result_large)
    run_engine(fetch_xrc_results_from_zocalo(zocalo, test_fgs_params))

    actual = x_ray_centre_event_handler.xray_centre_results
    expected = XRayCentreResult(
        centre_of_mass_mm=np.array([0.05, 0.15, 0.25]),
        bounding_box_mm=(
            np.array([0.15, 0.15, 0.15]),
            np.array([0.75, 0.75, 0.65]),
        ),
        max_count=105062,
        total_count=2387574,
        sample_id=12345,
    )
    assert actual and len(actual) == 1
    assert all(np.isclose(actual[0].centre_of_mass_mm, expected.centre_of_mass_mm))
    assert all(np.isclose(actual[0].bounding_box_mm[0], expected.bounding_box_mm[0]))
    assert all(np.isclose(actual[0].bounding_box_mm[1], expected.bounding_box_mm[1]))


def test_fetch_results_discards_results_below_threshold(
    test_fgs_params: SpecifiedThreeDGridScan,
    run_engine: RunEngine,
    zocalo: ZocaloResults,
):
    callback = XRayCentreEventHandler()
    run_engine.subscribe(callback)

    mock_zocalo_trigger(
        zocalo,
        TestData.test_result_medium
        + TestData.test_result_below_threshold
        + TestData.test_result_small,
    )
    run_engine(fetch_xrc_results_from_zocalo(zocalo, test_fgs_params))

    assert callback.xray_centre_results and len(callback.xray_centre_results) == 2
    assert [r.max_count for r in callback.xray_centre_results] == [50000, 1000]


def test_no_xtal_found_raises_exception(
    run_engine: RunEngine,
    test_fgs_params: SpecifiedThreeDGridScan,
    zocalo: ZocaloResults,
):
    mock_zocalo_trigger(zocalo, [])

    with pytest.raises(CrystalNotFoundError):
        run_engine(fetch_xrc_results_from_zocalo(zocalo, test_fgs_params))


def test_dummy_result_returned_when_no_xtal_and_commissioning_mode_enabled(
    run_engine: RunEngine,
    test_fgs_params: SpecifiedThreeDGridScan,
    fake_fgs_composite: FlyScanEssentialDevices,
    beamline_specific: BeamlineSpecificFGSFeatures,
    zocalo: ZocaloResults,
    baton_in_commissioning_mode,
):
    xrc_event_handler = XRayCentreEventHandler()
    run_engine.subscribe(xrc_event_handler)

    mock_zocalo_trigger(zocalo, [])

    run_engine(fetch_xrc_results_from_zocalo(zocalo, test_fgs_params))
    results = xrc_event_handler.xray_centre_results or []
    assert len(results) == 1
    result = results[0]
    assert result.sample_id == test_fgs_params.sample_id
    assert result.max_count == 10000
    assert result.total_count == 100000
    assert all(np.isclose(result.bounding_box_mm[0], [1.95, 0.95, 0.45]))
    assert all(np.isclose(result.bounding_box_mm[1], [2.05, 1.05, 0.55]))
    assert all(np.isclose(result.centre_of_mass_mm, [2.0, 1.0, 0.5]))


def test_zocalo_stage_wrapper(
    zocalo: ZocaloResults, sim_run_engine: RunEngineSimulator
):
    @zocalo_stage_decorator(zocalo)
    def test_plan():
        yield from bps.null()

    msgs = sim_run_engine.simulate_plan(test_plan())
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "stage"
        and msg.obj.name == "zocalo"
        and msg.kwargs["group"] == ZOCALO_STAGE_GROUP,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "unstage" and msg.obj.name == "zocalo",
    )



from mx_bluesky.common.utils.utils import (
    fix_transmission_and_exposure_time_for_current_wavelength,
)

ASSUMED_WAVELENGTH = 0.95373


def test_fix_transmission_and_exposure_time_does_nothing_given_assumed_energy():
    transmission = 0.55
    exposure_time_s = 0.03
    assert fix_transmission_and_exposure_time_for_current_wavelength(
        ASSUMED_WAVELENGTH, ASSUMED_WAVELENGTH, transmission, exposure_time_s
    ) == (transmission, exposure_time_s)


def test_fix_transmission_and_exposure_time_different_wavelength_reduces_transmission_only():
    transmission = 0.55
    exposure_time_s = 0.03
    wavelength_to_use = ASSUMED_WAVELENGTH * 2
    expected_transmission = transmission * (
        (ASSUMED_WAVELENGTH / wavelength_to_use) ** 2
    )
    assert fix_transmission_and_exposure_time_for_current_wavelength(
        wavelength_to_use, ASSUMED_WAVELENGTH, transmission, exposure_time_s
    ) == (expected_transmission, exposure_time_s)


def test_fix_transmission_and_exposure_time_low_wavelength_lowers_exposure_time_if_transmission_is_maxed():
    transmission = 0.95
    exposure_time_s = 0.03
    wavelength_to_use = ASSUMED_WAVELENGTH / 2
    expected_transmission = 1
    expected_exposure_time = round(
        transmission
        * ((ASSUMED_WAVELENGTH / wavelength_to_use) ** 2)
        * exposure_time_s,
        3,
    )
    assert fix_transmission_and_exposure_time_for_current_wavelength(
        wavelength_to_use, ASSUMED_WAVELENGTH, transmission, exposure_time_s
    ) == (expected_transmission, expected_exposure_time)


def test_fix_transmission_and_exposure_time_rounds_exposure_time_to_ms():
    exposure_time_s = 0.234873469
    transmission = 1
    assert fix_transmission_and_exposure_time_for_current_wavelength(
        ASSUMED_WAVELENGTH, ASSUMED_WAVELENGTH, transmission, exposure_time_s
    ) == (transmission, round(exposure_time_s, 3))



from collections.abc import Callable, Generator, Sequence
from copy import deepcopy
from functools import partial
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import bluesky.plan_stubs as bps
import ispyb.sqlalchemy
import numpy
import pytest
import pytest_asyncio
from dodal.beamlines import i03
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.beamlines.i03 import Beamstop
from dodal.devices.beamlines.i03.dcm import DCM
from dodal.devices.beamsize.beamsize import BeamsizeBase
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.flux import Flux
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron, SynchrotronMode
from dodal.devices.thawer import Thawer
from dodal.devices.undulator import UndulatorInKeV
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from ispyb.sqlalchemy import (
    BLSample,
    DataCollection,
    DataCollectionGroup,
    GridInfo,
    Position,
)
from ophyd_async.core import (
    AsyncStatus,
    callback_on_mock_put,
    completed_status,
    set_mock_value,
)
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from workflows.recipe import RecipeWrapper

from mx_bluesky.common.external_interaction.ispyb.ispyb_store import StoreInIspyb
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.common.parameters.rotation import (
    RotationScan,
)
from mx_bluesky.common.utils.utils import convert_angstrom_to_ev
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
    HyperionGridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan

from ....conftest import (
    TEST_RESULT_MEDIUM,
    SimConstants,
    pin_tip_edge_data,
    raw_params_from_file,
)


def get_current_datacollection_comment(session: Callable, dcid: int) -> str:
    """Read the 'comments' field from the given datacollection id's ISPyB entry.
    Returns an empty string if the comment is not yet initialised.
    """
    try:
        with session() as _session:
            query = _session.query(DataCollection).filter(
                DataCollection.dataCollectionId == dcid
            )
            current_comment: str = query.first().comments
    except Exception:
        current_comment = ""
    return current_comment


def get_datacollections(session: Callable, dcg_id: int) -> Sequence[int]:
    with session.begin() as _session:  # type: ignore
        query = _session.query(DataCollection.dataCollectionId).filter(
            DataCollection.dataCollectionGroupId == dcg_id
        )
        return [row[0] for row in query.all()]


def get_current_datacollection_attribute(
    session: Callable, dcid: int, attr: str
) -> str:
    """Read the specified field 'attr' from the given datacollection id's ISPyB entry.
    Returns an empty string if the attribute is not found.
    """
    try:
        with session() as _session:
            query = _session.query(DataCollection).filter(
                DataCollection.dataCollectionId == dcid
            )
            first_result = query.first()
            data: str = getattr(first_result, attr)
    except Exception:
        data = ""
    return data


def get_current_datacollection_grid_attribute(
    session: Callable, grid_id: int, attr: str
) -> Any:
    with session() as _session:
        query = _session.query(GridInfo).filter(GridInfo.gridInfoId == grid_id)
        first_result = query.first()
        return getattr(first_result, attr)


def get_current_position_attribute(
    session: Callable, position_id: int, attr: str
) -> Any:
    with session() as _session:
        query = _session.query(Position).filter(Position.positionId == position_id)
        first_result = query.first()
        if first_result is None:
            return None
        return getattr(first_result, attr)


def get_current_datacollectiongroup_attribute(
    session: Callable, dcg_id: int, attr: str
):
    with session() as _session:
        query = _session.query(DataCollectionGroup).filter(
            DataCollectionGroup.dataCollectionGroupId == dcg_id
        )
        first_result = query.first()
        return getattr(first_result, attr)


def get_blsample(session: Callable, bl_sample_id: int) -> BLSample:
    with session() as _session:
        query = _session.query(BLSample).filter(BLSample.blSampleId == bl_sample_id)
        return query.first()


@pytest.fixture
def sqlalchemy_sessionmaker(ispyb_config_path) -> sessionmaker:
    url = ispyb.sqlalchemy.url(ispyb_config_path)
    engine = create_engine(url, connect_args={"use_pure": True})
    return sessionmaker(engine)


@pytest.fixture
def fetch_comment(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollection_comment, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_ids_for_group_id(
    sqlalchemy_sessionmaker,
) -> Callable[[int], Sequence]:
    return partial(get_datacollections, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollection_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_grid_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollection_grid_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollection_position_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_position_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_datacollectiongroup_attribute(sqlalchemy_sessionmaker) -> Callable:
    return partial(get_current_datacollectiongroup_attribute, sqlalchemy_sessionmaker)


@pytest.fixture
def fetch_blsample(sqlalchemy_sessionmaker) -> Callable[[int], BLSample]:
    return partial(get_blsample, sqlalchemy_sessionmaker)


@pytest.fixture
def dummy_params(tmp_path):
    params_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/test_gridscan_param_defaults.json",
        tmp_path,
    )
    dummy_params = HyperionSpecifiedThreeDGridScan(**params_dict)
    dummy_params.visit = SimConstants.ST_VISIT
    dummy_params.sample_id = SimConstants.ST_SAMPLE_ID
    return dummy_params


@pytest.fixture
def dummy_ispyb(ispyb_config_path, dummy_params) -> StoreInIspyb:
    return StoreInIspyb(ispyb_config_path)


@pytest_asyncio.fixture
async def zocalo_for_fake_zocalo(zocalo_env) -> ZocaloResults:
    """
    This attempts to connect to a fake zocalo via rabbitmq
    """
    zd = ZocaloResults("zocalo")
    zd.timeout_s = 10
    await zd.connect()
    return zd


@pytest.fixture
def zocalo_for_system_test() -> Generator[ZocaloResults, None, None]:
    zocalo = i03.zocalo.build(connect_immediately=True, mock=True)
    zocalo.timeout_s = 10
    old_zocalo_trigger = zocalo.trigger
    zocalo.my_zocalo_result = deepcopy(TEST_RESULT_MEDIUM)
    zocalo.my_zocalo_result[0]["sample_id"] = SimConstants.ST_SAMPLE_ID  # type: ignore

    @AsyncStatus.wrap
    async def mock_zocalo_complete():
        fake_recipe_wrapper = MagicMock(spec=RecipeWrapper)
        fake_recipe_wrapper.recipe_step = {
            "parameters": {"dcid": 1234, "dcgid": 123, "gpu": True}
        }
        message = {
            "results": zocalo.my_zocalo_result  # type: ignore
        }
        header = {}
        zocalo.my_callback(fake_recipe_wrapper, header, message)  # type: ignore
        await old_zocalo_trigger()

    def mock_worfklow_subscribe(transport, channel, callback, **kwargs):
        if channel == "xrc.i03":
            zocalo.my_callback = callback

    with (
        patch("dodal.devices.zocalo.zocalo_results.workflows") as workflows,
        patch("dodal.devices.zocalo.zocalo_results._get_zocalo_connection"),
    ):
        workflows.recipe.wrap_subscribe.side_effect = mock_worfklow_subscribe
        with patch.object(zocalo, "trigger", side_effect=mock_zocalo_complete):
            yield zocalo


@pytest.fixture
def grid_detect_then_xray_centre_composite(
    fast_grid_scan,
    backlight,
    beamstop_phase1,
    smargon,
    undulator_for_system_test,
    synchrotron,
    s4_slit_gaps,
    attenuator,
    xbpm_feedback,
    detector_motion,
    zocalo_for_system_test,
    aperture_scatterguard,
    zebra,
    eiger,
    robot,
    oav_for_system_test,
    dcm,
    flux,
    ophyd_pin_tip_detection,
    sample_shutter,
    panda,
    panda_fast_grid_scan,
    request,
    beamsize: BeamsizeBase,
):
    composite = HyperionGridDetectThenXRayCentreComposite(
        zebra_fast_grid_scan=fast_grid_scan,
        pin_tip_detection=ophyd_pin_tip_detection,
        backlight=backlight,
        beamstop=beamstop_phase1,
        panda_fast_grid_scan=panda_fast_grid_scan,
        gonio=smargon,
        undulator=undulator_for_system_test,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        attenuator=attenuator,
        xbpm_feedback=xbpm_feedback,
        detector_motion=detector_motion,
        zocalo=zocalo_for_system_test,
        aperture_scatterguard=aperture_scatterguard,
        zebra=zebra,
        eiger=eiger,
        panda=panda,
        robot=robot,
        oav=oav_for_system_test,
        dcm=dcm,
        flux=flux,
        sample_shutter=sample_shutter,
        beamsize=beamsize,
    )

    def default_edge_generator():
        while True:
            yield pin_tip_edge_data()

    edge_data_generator = default_edge_generator()
    if param := getattr(request, "param", None):
        edge_data_generator = param()

    @AsyncStatus.wrap
    async def mock_pin_tip_detect():
        tip_x_px, tip_y_px, top_edge_array, bottom_edge_array = next(
            edge_data_generator
        )
        set_mock_value(
            ophyd_pin_tip_detection.triggered_top_edge,
            top_edge_array,
        )

        set_mock_value(
            ophyd_pin_tip_detection.triggered_bottom_edge,
            bottom_edge_array,
        )
        set_mock_value(
            zocalo_for_system_test.bounding_box,
            numpy.array([[10, 10, 10]], dtype=numpy.uint64),
        )
        set_mock_value(
            ophyd_pin_tip_detection.triggered_tip, numpy.array([tip_x_px, tip_y_px])
        )

    with (
        patch.object(eiger, "wait_on_arming_if_started"),
        # xsize, ysize will always be wrong since computed as 0 before we get here
        # patch up load_microns_per_pixel connect to receive non-zero values
        patch.object(
            ophyd_pin_tip_detection, "trigger", side_effect=mock_pin_tip_detect
        ),
        patch.object(fast_grid_scan, "kickoff", side_effect=lambda: completed_status()),
        patch.object(
            fast_grid_scan, "complete", side_effect=lambda: completed_status()
        ),
    ):
        yield composite


@pytest.fixture
def fgs_composite_for_fake_zocalo(
    hyperion_flyscan_xrc_composite: HyperionFlyScanXRayCentreComposite,
    zocalo_for_fake_zocalo: ZocaloResults,
) -> HyperionFlyScanXRayCentreComposite:
    set_mock_value(
        hyperion_flyscan_xrc_composite.aperture_scatterguard.aperture.z.user_setpoint, 2
    )
    hyperion_flyscan_xrc_composite.eiger.unstage = MagicMock(
        side_effect=lambda: completed_status()
    )  # type: ignore
    hyperion_flyscan_xrc_composite.gonio.stub_offsets.set = MagicMock(
        side_effect=lambda _: completed_status()
    )  # type: ignore
    callback_on_mock_put(
        hyperion_flyscan_xrc_composite.zebra_fast_grid_scan.run_cmd,
        lambda *args, **kwargs: set_mock_value(
            hyperion_flyscan_xrc_composite.zebra_fast_grid_scan.status, 1
        ),
    )
    hyperion_flyscan_xrc_composite.zebra_fast_grid_scan.complete = MagicMock(
        side_effect=lambda: completed_status()
    )
    hyperion_flyscan_xrc_composite.zocalo = zocalo_for_fake_zocalo
    return hyperion_flyscan_xrc_composite


@pytest.fixture
def pin_tip_no_pin_found(ophyd_pin_tip_detection):
    @AsyncStatus.wrap
    async def no_pin_tip_found():
        set_mock_value(
            ophyd_pin_tip_detection.triggered_tip, PinTipDetection.INVALID_POSITION
        )

        set_mock_value(
            ophyd_pin_tip_detection.triggered_top_edge,
            numpy.array([]),
        )

        set_mock_value(
            ophyd_pin_tip_detection.triggered_bottom_edge,
            numpy.array([]),
        )

    with patch.object(ophyd_pin_tip_detection, "trigger", side_effect=no_pin_tip_found):
        yield ophyd_pin_tip_detection


@pytest.fixture
def params_for_rotation_scan(
    test_rotation_params: RotationScan,
) -> RotationScan:
    test_rotation_params.rotation_increment_deg = 0.27
    test_rotation_params.exposure_time_s = 0.023
    test_rotation_params.detector_params.expected_energy_ev = 0.71
    test_rotation_params.visit = SimConstants.ST_VISIT
    test_rotation_params.rotation_scans[0].sample_id = SimConstants.ST_SAMPLE_ID
    return test_rotation_params


@pytest.fixture
def composite_for_rotation_scan(
    beamstop_phase1: Beamstop,
    eiger: EigerDetector,
    smargon: Smargon,
    zebra: Zebra,
    detector_motion: DetectorMotion,
    backlight: Backlight,
    attenuator: BinaryFilterAttenuator,
    flux: Flux,
    undulator_for_system_test: UndulatorInKeV,
    aperture_scatterguard: ApertureScatterguard,
    synchrotron: Synchrotron,
    s4_slit_gaps: S4SlitGaps,
    dcm: DCM,
    robot: BartRobot,
    oav_for_system_test: OAV,
    sample_shutter: ZebraShutter,
    xbpm_feedback: XBPMFeedback,
    thawer: Thawer,
    beamsize: BeamsizeBase,
):
    set_mock_value(smargon.omega.max_velocity, 131)
    oav_for_system_test.zoom_controller.level.describe = AsyncMock(
        return_value={"level": {"choices": ["1.0x", "5.0x", "7.5x"]}}
    )

    fake_create_rotation_devices = RotationScanComposite(
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_phase1,
        dcm=dcm,
        detector_motion=detector_motion,
        eiger=eiger,
        flux=flux,
        gonio=smargon,
        undulator=undulator_for_system_test,
        aperture_scatterguard=aperture_scatterguard,
        synchrotron=synchrotron,
        s4_slit_gaps=s4_slit_gaps,
        zebra=zebra,
        robot=robot,
        oav=oav_for_system_test,
        sample_shutter=sample_shutter,
        xbpm_feedback=xbpm_feedback,
        thawer=thawer,
        beamsize=beamsize,
    )

    energy_ev = convert_angstrom_to_ev(0.71)
    set_mock_value(
        fake_create_rotation_devices.dcm.energy_in_keV.user_readback,
        energy_ev / 1000,  # pyright: ignore
    )
    set_mock_value(
        fake_create_rotation_devices.synchrotron.synchrotron_mode,
        SynchrotronMode.USER,
    )
    set_mock_value(
        fake_create_rotation_devices.synchrotron.top_up_start_countdown,
        -1,
    )
    set_mock_value(fake_create_rotation_devices.s4_slit_gaps.xgap.user_readback, 0.123)
    set_mock_value(fake_create_rotation_devices.s4_slit_gaps.ygap.user_readback, 0.234)

    yield fake_create_rotation_devices


@pytest.fixture
def fake_grid_snapshot_plan():
    def plan(smargon, oav):
        for omega in [-90, 0]:
            yield from bps.mv(smargon.omega, omega)
            yield from bps.create(DocDescriptorNames.OAV_GRID_SNAPSHOT_TRIGGERED)

            yield from bps.read(oav)
            yield from bps.read(smargon)
            yield from bps.save()

    return plan



import dataclasses
import json
import re
from functools import partial

import pytest
import responses
from requests import PreparedRequest

TEST_DATA_COLLECTION_IDS = (12, 13)
TEST_DATA_COLLECTION_GROUP_ID = 34
TEST_GRID_INFO_IDS = (56, 57)

URL_PREFIX = "(http://[^/]+|)"
DCGS_RE = re.compile(URL_PREFIX + r"/proposals/([a-z]+\d+)/sessions/(\d+)/data-groups$")
DCG_RE = re.compile(URL_PREFIX + r"/data-groups/(\d+)$")
DCS_RE = re.compile(URL_PREFIX + r"/data-groups/(\d+)/data-collections$")
DC_RE = re.compile(URL_PREFIX + r"/data-collections/(\d+)$")
DC_COMMENT_RE = re.compile(URL_PREFIX + r"/data-collections/(\d+)\?appendComment=true$")
POSITION_RE = re.compile(URL_PREFIX + r"/data-collections/(\d+)/position$")
GRID_RE = re.compile(URL_PREFIX + r"/data-collections/(\d+)/grids$")


@dataclasses.dataclass
class ExpeyeDCRequestInfo:
    dcid: int
    url: str
    body: dict


def _mock_ispyb_conn(dcgid, dcids, giids):
    def create_or_update_dcg_response(status, request):
        return (
            status,
            {},
            json.dumps(
                {
                    "dataCollectionGroupId": dcgid,
                }
            ),
        )

    it_dc_id = iter(dcids)

    def create_dc_response(request):
        requested_dcg_id = int(DCS_RE.match(request.path_url)[2])  # type: ignore
        assert requested_dcg_id == dcgid
        return (
            201,
            {},
            json.dumps(
                {
                    "dataCollectionId": next(it_dc_id),  # type: ignore
                    "dataCollectionGroupId": dcgid,
                }
            ),
        )

    def update_dc_response(pattern, request):
        requested_dc_id = int(pattern.match(request.path_url)[2])  # type: ignore
        assert requested_dc_id in dcids
        return (
            200,
            {},
            json.dumps(
                {"dataCollectionId": requested_dc_id, "dataCollectionGroupId": dcgid}
            ),
        )

    def create_position_response(request):
        requested_dc_id = int(POSITION_RE.match(request.path_url)[2])  # type: ignore
        assert requested_dc_id in dcids
        return (201, {}, json.dumps({}))

    it_grid_info_id = iter(giids)

    def create_grid_response(request):
        requested_dc_id = int(GRID_RE.match(request.path_url)[2])  # type: ignore
        assert requested_dc_id in dcids
        return (201, {}, json.dumps({"gridInfoId": next(it_grid_info_id)}))  # type: ignore

    class ExpeyeRequestsUtil:
        def __init__(self, mock_req: responses.RequestsMock):
            self.mock_req = mock_req
            for pattern, callback in {
                DCGS_RE: partial(create_or_update_dcg_response, 201),
                DCS_RE: create_dc_response,
                POSITION_RE: create_position_response,
                GRID_RE: create_grid_response,
            }.items():
                self.mock_req.add_callback(
                    responses.POST,
                    pattern,
                    callback=callback,
                    content_type="application/json",
                )

            for pattern, callback in {
                DC_RE: partial(update_dc_response, DC_RE),
                DC_COMMENT_RE: partial(update_dc_response, DC_COMMENT_RE),
                DCG_RE: partial(create_or_update_dcg_response, 200),
            }.items():
                self.mock_req.add_callback(
                    responses.PATCH,
                    pattern,
                    callback=callback,
                    content_type="application/json",
                )

        def calls_for(self, pattern: str | re.Pattern):
            if not isinstance(pattern, re.Pattern):
                return [c for c in self.mock_req.calls if c.request.url == pattern]
            else:
                return [c for c in self.mock_req.calls if pattern.match(c.request.url)]

        def dc_calls_for(self, pattern: re.Pattern):
            return [
                ExpeyeDCRequestInfo(
                    dcid=int(pattern.match(c.request.url)[2]),  # type: ignore
                    url=c.request.url,  # type: ignore
                    body=json.loads(c.request.body),  # type: ignore
                )
                for c in self.calls_for(pattern)
            ]

        def match(self, req: PreparedRequest, pattern: re.Pattern, idx: int) -> str:
            matcher = pattern.match(req.url)
            assert matcher
            return matcher[idx]

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rq_mock:
        yield ExpeyeRequestsUtil(rq_mock)


@pytest.fixture
def mock_ispyb_conn(request):
    dcg_id = getattr(request, "param", {}).get("dcg_id", TEST_DATA_COLLECTION_GROUP_ID)
    yield from _mock_ispyb_conn(
        dcg_id,
        TEST_DATA_COLLECTION_IDS,
        TEST_GRID_INFO_IDS,
    )


@pytest.fixture
def mock_ispyb_conn_multiscan():
    yield from _mock_ispyb_conn(
        TEST_DATA_COLLECTION_GROUP_ID,
        list(range(12, 24)),
        list(range(56, 68)),
    )



from __future__ import annotations

from collections.abc import Callable
from functools import partial
from pathlib import Path

import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.fast_grid_scan import (
    set_fast_grid_scan_params,
)

from mx_bluesky.common.device_setup_plans.setup_zebra_and_shutter import (
    setup_zebra_for_gridscan,
    tidy_up_zebra_after_gridscan,
)
from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    construct_beamline_specific_fast_gridscan_features,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.device_setup_plans.setup_panda import (
    disarm_panda_for_gridscan,
    set_panda_directory,
    setup_panda_for_flyscan,
)
from mx_bluesky.hyperion.device_setup_plans.setup_zebra import (
    setup_zebra_for_panda_flyscan,
)
from mx_bluesky.hyperion.external_interaction.config_server import (
    get_hyperion_feature_settings,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import HyperionSpecifiedThreeDGridScan


class SmargonSpeedError(Exception):
    pass


def construct_hyperion_specific_features(
    xrc_composite: HyperionFlyScanXRayCentreComposite,
    xrc_parameters: HyperionSpecifiedThreeDGridScan,
):
    """
    Get all the information needed to do the Hyperion-specific parts of the XRC flyscan.
    """
    signals_to_read_pre_flyscan = [
        xrc_composite.undulator.current_gap,
        xrc_composite.synchrotron.synchrotron_mode,
        xrc_composite.s4_slit_gaps,
        xrc_composite.gonio,
        xrc_composite.dcm.energy_in_keV,
    ]

    signals_to_read_during_collection = [
        xrc_composite.aperture_scatterguard,
        xrc_composite.attenuator.actual_transmission,
        xrc_composite.flux.flux_reading,
        xrc_composite.dcm.energy_in_keV,
        xrc_composite.eiger.bit_depth,
        xrc_composite.beamsize,
        xrc_composite.eiger.cam.roi_mode,
        xrc_composite.eiger.ispyb_detector_id,
    ]

    setup_trigger_plan: Callable[..., MsgGenerator]

    if get_hyperion_feature_settings().USE_PANDA_FOR_GRIDSCAN:
        setup_trigger_plan = _panda_triggering_setup
        tidy_plan = partial(_panda_tidy, xrc_composite)
        set_flyscan_params_plan = partial(
            set_fast_grid_scan_params,
            xrc_composite.panda_fast_grid_scan,
            xrc_parameters.panda_fast_gridscan_params,
        )
        fgs_motors = xrc_composite.panda_fast_grid_scan

    else:
        setup_trigger_plan = partial(
            setup_zebra_for_gridscan,
        )
        tidy_plan = partial(
            tidy_up_zebra_after_gridscan,
            xrc_composite.zebra,
            xrc_composite.sample_shutter,
            group="flyscan_zebra_tidy",
            wait=True,
        )
        set_flyscan_params_plan = partial(
            set_fast_grid_scan_params,
            xrc_composite.zebra_fast_grid_scan,
            xrc_parameters.fast_gridscan_params,
        )
        fgs_motors = xrc_composite.zebra_fast_grid_scan
    return construct_beamline_specific_fast_gridscan_features(
        setup_trigger_plan,
        tidy_plan,
        set_flyscan_params_plan,
        fgs_motors,
        signals_to_read_pre_flyscan,
        signals_to_read_during_collection,  # type: ignore # until https://github.com/DiamondLightSource/mx-bluesky/issues/1076
    )


def _panda_tidy(xrc_composite: HyperionFlyScanXRayCentreComposite):
    group = "panda_flyscan_tidy"
    LOGGER.info("Disabling panda blocks")
    yield from disarm_panda_for_gridscan(xrc_composite.panda, group)
    yield from tidy_up_zebra_after_gridscan(
        xrc_composite.zebra, xrc_composite.sample_shutter, group=group, wait=False
    )
    yield from bps.unstage(xrc_composite.panda, group=group)
    yield from bps.wait(group, timeout=10)


def _panda_triggering_setup(
    xrc_composite: HyperionFlyScanXRayCentreComposite,
    parameters: HyperionSpecifiedThreeDGridScan,
) -> MsgGenerator:
    LOGGER.info("Setting up Panda for flyscan")

    run_up_distance_mm = yield from bps.rd(
        xrc_composite.panda_fast_grid_scan.run_up_distance_mm
    )

    detector_deadtime_s = 1e-4  # This value was empirically found to be safer than the documented deadtime in the Eiger manual

    time_between_x_steps_ms = (detector_deadtime_s + parameters.exposure_time_s) * 1e3

    smargon_speed_limit_mm_per_s = yield from bps.rd(xrc_composite.gonio.x.max_velocity)

    sample_velocity_mm_per_s = (
        parameters.panda_fast_gridscan_params.x_step_size_mm
        * 1e3
        / time_between_x_steps_ms
    )
    if sample_velocity_mm_per_s > smargon_speed_limit_mm_per_s:
        raise SmargonSpeedError(
            f"Smargon speed was calculated from x step size\
            {parameters.panda_fast_gridscan_params.x_step_size_mm}mm and\
            time_between_x_steps_ms {time_between_x_steps_ms} as\
            {sample_velocity_mm_per_s}mm/s. The smargon's speed limit is\
            {smargon_speed_limit_mm_per_s}mm/s."
        )
    else:
        LOGGER.info(
            f"Panda grid scan: Smargon speed set to {sample_velocity_mm_per_s} mm/s"
            f" and using a run-up distance of {run_up_distance_mm}"
        )

    yield from bps.mv(
        xrc_composite.panda_fast_grid_scan.time_between_x_steps_ms,
        time_between_x_steps_ms,
    )

    directory_provider_root = Path(parameters.storage_directory)
    yield from set_panda_directory(directory_provider_root)

    yield from setup_panda_for_flyscan(
        xrc_composite.panda,
        parameters.panda_fast_gridscan_params,
        xrc_composite.gonio,
        parameters.exposure_time_s,
        time_between_x_steps_ms,
        sample_velocity_mm_per_s,
    )

    LOGGER.info("Setting up Zebra for panda flyscan")
    yield from setup_zebra_for_panda_flyscan(
        xrc_composite.zebra, xrc_composite.sample_shutter, wait=True
    )
